#    Copyright 2015 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Cinder common internal object model"""

from collections import abc
import contextlib
import datetime

from oslo_log import log as logging
from oslo_utils import versionutils
from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from cinder import db
from cinder import exception
from cinder.i18n import _
from cinder import objects


LOG = logging.getLogger('object')
obj_make_list = base.obj_make_list


class CinderObjectVersionsHistory(dict):
    """Helper class that maintains objects version history.

    Current state of object versions is aggregated in a single version number
    that explicitly identifies a set of object versions. That way a service
    is able to report what objects it supports using a single string and all
    the newer services will know exactly what that mean for a single object.
    """

    def __init__(self):
        super(CinderObjectVersionsHistory, self).__init__()
        # NOTE(dulek): This is our pre-history and a starting point - Liberty.
        # We want Mitaka to be able to talk to Liberty services, so we need to
        # handle backporting to these objects versions (although I don't expect
        # we've made a lot of incompatible changes inside the objects).
        #
        # If an object doesn't exist in Liberty, RPC API compatibility layer
        # shouldn't send it or convert it to a dictionary.
        #
        # Please note that we do not need to add similar entires for each
        # release. Liberty is here just for historical reasons.
        self.versions = ['1.38']
        self['1.38'] = {
            'Backup': '1.7',
            'BackupDeviceInfo': '1.0',
            'BackupImport': '1.7',
            'BackupList': '1.0',
            'CleanupRequest': '1.0',
            'CGSnapshot': '1.1',
            'CGSnapshotList': '1.0',
            'Cluster': '1.1',
            'ClusterList': '1.0',
            'ConsistencyGroup': '1.4',
            'ConsistencyGroupList': '1.1',
            'Group': '1.2',
            'GroupList': '1.0',
            'GroupSnapshot': '1.0',
            'GroupSnapshotList': '1.0',
            'GroupType': '1.0',
            'GroupTypeList': '1.0',
            'LogLevel': '1.0',
            'LogLevelList': '1.0',
            'ManageableSnapshot': '1.0',
            'ManageableSnapshotList': '1.0',
            'ManageableVolume': '1.0',
            'ManageableVolumeList': '1.0',
            'QualityOfServiceSpecs': '1.0',
            'QualityOfServiceSpecsList': '1.0',
            'RequestSpec': '1.5',
            'Service': '1.6',
            'ServiceList': '1.1',
            'Snapshot': '1.5',
            'SnapshotList': '1.0',
            'Volume': '1.8',
            'VolumeAttachment': '1.3',
            'VolumeAttachmentList': '1.1',
            'VolumeList': '1.1',
            'VolumeProperties': '1.1',
            'VolumeType': '1.3',
            'VolumeTypeList': '1.1',
        }

    def get_current(self):
        return self.versions[-1]

    def get_current_versions(self):
        return self[self.get_current()]

    def add(self, ver, updates):
        if ver in self.versions:
            msg = 'Version %s already exists in history.' % ver
            raise exception.ProgrammingError(reason=msg)

        self[ver] = self[self.get_current()].copy()
        self.versions.append(ver)
        self[ver].update(updates)


OBJ_VERSIONS = CinderObjectVersionsHistory()
# NOTE(dulek): You should add a new version here each time you bump a version
# of any object. As a second parameter you need to specify only what changed.

# On each release we should drop backward compatibility with -2 release, since
# rolling upgrades only needs to support compatibility with previous release.
# So if we are in N release we can remove history from L and earlier.
# Example of how to keep track of this:
#     # TODO: (T release) remove up to next TODO (was added in R release) and
#     #       update CinderObjectVersionsHistory
#     OBJ_VERSIONS.add('1.34', {'VolumeAttachment': '1.3'})
#     OBJ_VERSIONS.add('1.35', {'Backup': '1.6', 'BackupImport': '1.6'})
#
#     # TODO: (U release) remove up to next TODO (was added in S release) and
#     #       update CinderObjectVersionsHistory
#     OBJ_VERSIONS.add('1.36', {'RequestSpec': '1.4'})
#     OBJ_VERSIONS.add('1.37', {'RequestSpec': '1.5'})
#     OBJ_VERSIONS.add('1.38', {'Backup': '1.7', 'BackupImport': '1.7'})
# When we reach T release we remove versions 1.34 and 1.35 and update __init__
# method in CinderObjectVerseionsHistory to bump VolumeAttachment to 1.3,
# Backup to 1.6 and BackupImport to 1.6, and changing the versions list to
# '1.35' and the self['<versioname>'] = { to self['1.35'] = {


# TODO: (Z release) remove up to next TODO  and update
#       CinderObjectVersionsHistory (was added in X release)
OBJ_VERSIONS.add('1.39', {'Volume': '1.9', 'Snapshot': '1.6'})


class CinderObjectRegistry(base.VersionedObjectRegistry):
    def registration_hook(self, cls, index):
        """Hook called when registering a class.

        This method takes care of adding the class to cinder.objects namespace.

        Should registering class have a method called cinder_ovo_cls_init it
        will be called to support class initialization.  This is convenient
        for all persistent classes that need to register their models.
        """
        setattr(objects, cls.obj_name(), cls)

        # If registering class has a callable initialization method, call it.
        if isinstance(getattr(cls, 'cinder_ovo_cls_init', None),
                      abc.Callable):
            cls.cinder_ovo_cls_init()


class CinderObject(base.VersionedObject):
    # NOTE(thangp): OBJ_PROJECT_NAMESPACE needs to be set so that nova,
    # cinder, and other objects can exist on the same bus and be distinguished
    # from one another.
    OBJ_PROJECT_NAMESPACE = 'cinder'

    def cinder_obj_get_changes(self):
        """Returns a dict of changed fields with tz unaware datetimes.

        Any timezone aware datetime field will be converted to UTC timezone
        and returned as timezone unaware datetime.

        This will allow us to pass these fields directly to a db update
        method as they can't have timezone information.
        """
        # Get dirtied/changed fields
        changes = self.obj_get_changes()

        # Look for datetime objects that contain timezone information
        for k, v in changes.items():
            if isinstance(v, datetime.datetime) and v.tzinfo:
                # Remove timezone information and adjust the time according to
                # the timezone information's offset.
                changes[k] = v.replace(tzinfo=None) - v.utcoffset()

        # Return modified dict
        return changes

    def obj_make_compatible(self, primitive, target_version):
        _log_backport(self, target_version)
        super(CinderObject, self).obj_make_compatible(primitive,
                                                      target_version)

    def __contains__(self, name):
        # We're using obj_extra_fields to provide aliases for some fields while
        # in transition period. This override is to make these aliases pass
        # "'foo' in obj" tests.
        return name in self.obj_extra_fields or super(CinderObject,
                                                      self).__contains__(name)


class CinderObjectDictCompat(base.VersionedObjectDictCompat):
    """Mix-in to provide dictionary key access compat.

    If an object needs to support attribute access using
    dictionary items instead of object attributes, inherit
    from this class. This should only be used as a temporary
    measure until all callers are converted to use modern
    attribute access.

    NOTE(berrange) This class will eventually be deleted.
    """

    def get(self, key, value=base._NotSpecifiedSentinel):
        """For backwards-compatibility with dict-based objects.

        NOTE(danms): May be removed in the future.
        """
        if key not in self.obj_fields:
            # NOTE(jdg): There are a number of places where we rely on the
            # old dictionary version and do a get(xxx, None).
            # The following preserves that compatibility but in
            # the future we'll remove this shim altogether so don't
            # rely on it.
            LOG.debug('Cinder object %(object_name)s has no '
                      'attribute named: %(attribute_name)s',
                      {'object_name': self.__class__.__name__,
                       'attribute_name': key})
            return None
        if (value != base._NotSpecifiedSentinel and
                key not in self.obj_extra_fields and
                not self.obj_attr_is_set(key)):
            return value
        else:
            try:
                return getattr(self, key)
            except (exception.ObjectActionError, NotImplementedError):
                # Exception when haven't set a value for non-lazy
                # loadable attribute, but to mimic typical dict 'get'
                # behavior we should still return None
                return None


class CinderPersistentObject(object):
    """Mixin class for Persistent objects.

    This adds the fields that we use in common for all persistent objects.
    """
    OPTIONAL_FIELDS = []

    Not = db.Not
    Case = db.Case

    fields = {
        'created_at': fields.DateTimeField(nullable=True),
        'updated_at': fields.DateTimeField(nullable=True),
        'deleted_at': fields.DateTimeField(nullable=True),
        'deleted': fields.BooleanField(default=False,
                                       nullable=True),
    }

    @classmethod
    def cinder_ovo_cls_init(cls):
        """This method is called on OVO registration and sets the DB model."""
        # Persistent Versioned Objects Classes should have a DB model, and if
        # they don't, then we have a problem and we must raise an exception on
        # registration.
        try:
            cls.model = db.get_model_for_versioned_object(cls)
        except (ImportError, AttributeError):
            msg = _("Couldn't find ORM model for Persistent Versioned "
                    "Object %s.") % cls.obj_name()
            LOG.exception("Failed to initialize object.")
            raise exception.ProgrammingError(reason=msg)

    @contextlib.contextmanager
    def obj_as_admin(self):
        """Context manager to make an object call as an admin.

        This temporarily modifies the context embedded in an object to
        be elevated() and restores it after the call completes. Example
        usage:

           with obj.obj_as_admin():
               obj.save()
        """
        if self._context is None:
            raise exception.OrphanedObjectError(method='obj_as_admin',
                                                objtype=self.obj_name())

        original_context = self._context
        self._context = self._context.elevated()
        try:
            yield
        finally:
            self._context = original_context

    @contextlib.contextmanager
    def as_read_deleted(self, mode='yes'):
        """Context manager to make OVO with modified read deleted context.

        This temporarily modifies the context embedded in an object to
        have a different `read_deleted` parameter.

        Parameter mode accepts most of the same parameters as our `model_query`
        DB method.  We support 'yes', 'no', and 'only'.

        usage:

           with obj.as_read_deleted():
               obj.refresh()
           if obj.status = 'deleted':
               ...
        """
        if self._context is None:
            raise exception.OrphanedObjectError(method='as_read_deleted',
                                                objtype=self.obj_name())

        original_mode = self._context.read_deleted
        self._context.read_deleted = mode
        try:
            yield
        finally:
            self._context.read_deleted = original_mode

    @classmethod
    def _get_expected_attrs(cls, context, *args, **kwargs):
        return None

    @classmethod
    def get_by_id(cls, context, id, *args, **kwargs):
        # To get by id we need to have a model and for the model to
        # have an id field
        if 'id' not in cls.fields:
            msg = (_('VersionedObject %s cannot retrieve object by id.') %
                   (cls.obj_name()))
            raise NotImplementedError(msg)

        orm_obj = db.get_by_id(context, cls.model, id, *args, **kwargs)
        # We pass parameters because fields to expect may depend on them
        expected_attrs = cls._get_expected_attrs(context, *args, **kwargs)
        kargs = {}
        if expected_attrs:
            kargs = {'expected_attrs': expected_attrs}
        return cls._from_db_object(context, cls(context), orm_obj, **kargs)

    def update_single_status_where(self, new_status,
                                   expected_status, filters=()):
        values = {'status': new_status}
        expected_status = {'status': expected_status}
        return self.conditional_update(values, expected_status, filters)

    def conditional_update(self, values, expected_values=None, filters=(),
                           save_all=False, session=None, reflect_changes=True,
                           order=None):
        """Compare-and-swap update.

        A conditional object update that, unlike normal update, will SAVE the
        contents of the update to the DB.

        Update will only occur in the DB and the object if conditions are met.

        If no expected_values are passed in we will default to make sure that
        all fields have not been changed in the DB. Since we cannot know the
        original value in the DB for dirty fields in the object those will be
        excluded.

        We have 4 different condition types we can use in expected_values:
         - Equality:  {'status': 'available'}
         - Inequality: {'status': vol_obj.Not('deleting')}
         - In range: {'status': ['available', 'error']
         - Not in range: {'status': vol_obj.Not(['in-use', 'attaching'])

        Method accepts additional filters, which are basically anything that
        can be passed to a sqlalchemy query's filter method, for example:

        .. code-block:: python

         [~sql.exists().where(models.Volume.id == models.Snapshot.volume_id)]

        We can select values based on conditions using Case objects in the
        'values' argument. For example:

        .. code-block:: python

         has_snapshot_filter = sql.exists().where(
             models.Snapshot.volume_id == models.Volume.id)
         case_values = volume.Case([(has_snapshot_filter, 'has-snapshot')],
                                   else_='no-snapshot')
         volume.conditional_update({'status': case_values},
                                   {'status': 'available'}))

        And we can use DB fields using model class attribute for example to
        store previous status in the corresponding field even though we don't
        know which value is in the db from those we allowed:

        .. code-block:: python

         volume.conditional_update({'status': 'deleting',
                                    'previous_status': volume.model.status},
                                   {'status': ('available', 'error')})

        :param values: Dictionary of key-values to update in the DB.
        :param expected_values: Dictionary of conditions that must be met for
                                the update to be executed.
        :param filters: Iterable with additional filters
        :param save_all: Object may have changes that are not in the DB, this
                         will say whether we want those changes saved as well.
        :param session: Session to use for the update
        :param reflect_changes: If we want changes made in the database to be
                                reflected in the versioned object.  This may
                                mean in some cases that we have to reload the
                                object from the database.
        :param order: Specific order of fields in which to update the values
        :returns: number of db rows that were updated, which can be used as a
                  boolean, since it will be 0 if we couldn't update the DB and
                  1 if we could, because we are using unique index id.
        """
        if 'id' not in self.fields:
            msg = (_('VersionedObject %s does not support conditional update.')
                   % (self.obj_name()))
            raise NotImplementedError(msg)

        # If no conditions are set we will require object in DB to be unchanged
        if expected_values is None:
            changes = self.obj_what_changed()

            expected = {key: getattr(self, key)
                        for key in self.fields.keys()
                        if self.obj_attr_is_set(key) and key not in changes and
                        key not in self.OPTIONAL_FIELDS}
        else:
            # Set the id in expected_values to limit conditional update to only
            # change this object
            expected = expected_values.copy()
            expected['id'] = self.id

        # If we want to save any additional changes the object has besides the
        # ones referred in values
        if save_all:
            changes = self.cinder_obj_get_changes()
            changes.update(values)
            values = changes

        result = db.conditional_update(self._context, self.model, values,
                                       expected, filters, order=order)

        # If we were able to update the DB then we need to update this object
        # as well to reflect new DB contents and clear the object's dirty flags
        # for those fields.
        if result and reflect_changes:
            # If we have used a Case, a db field or an expression in values we
            # don't know which value was used, so we need to read the object
            # back from the DB
            if any(isinstance(v, self.Case) or db.is_orm_value(v)
                   for v in values.values()):
                # Read back object from DB
                obj = type(self).get_by_id(self._context, self.id)
                db_values = obj.obj_to_primitive()['versioned_object.data']
                # Only update fields were changes were requested
                values = {field: db_values[field]
                          for field, value in values.items()}

            # NOTE(geguileo): We don't use update method because our objects
            # will eventually move away from VersionedObjectDictCompat
            for key, value in values.items():
                setattr(self, key, value)
            self.obj_reset_changes(values.keys())
        return result

    def refresh(self):
        # To refresh we need to have a model and for the model to have an id
        # field
        if 'id' not in self.fields:
            msg = (_('VersionedObject %s cannot retrieve object by id.') %
                   (self.obj_name()))
            raise NotImplementedError(msg)

        current = self.get_by_id(self._context, self.id)

        # Copy contents retrieved from the DB into self
        my_data = vars(self)
        my_data.clear()
        my_data.update(vars(current))

    @classmethod
    def exists(cls, context, id_):
        return db.resource_exists(context, cls.model, id_)


class CinderComparableObject(base.ComparableVersionedObject):
    def __eq__(self, obj):
        if hasattr(obj, 'obj_to_primitive'):
            return self.obj_to_primitive() == obj.obj_to_primitive()
        return False

    def __ne__(self, other):
        return not self.__eq__(other)


class ObjectListBase(base.ObjectListBase):
    def obj_make_compatible(self, primitive, target_version):
        _log_backport(self, target_version)
        super(ObjectListBase, self).obj_make_compatible(primitive,
                                                        target_version)


class ClusteredObject(object):
    @property
    def service_topic_queue(self):
        return self.cluster_name or self.host

    @property
    def is_clustered(self):
        return bool(self.cluster_name)

    def assert_not_frozen(self):
        ctxt = self._context.elevated()
        if db.is_backend_frozen(ctxt, self.host, self.cluster_name):
            msg = _('Modification operations are not allowed on frozen '
                    'storage backends.')
            raise exception.InvalidInput(reason=msg)

    # The object's resource backend depends on whether it's clustered.
    resource_backend = service_topic_queue


class CinderObjectSerializer(base.VersionedObjectSerializer):
    OBJ_BASE_CLASS = CinderObject

    def __init__(self, version_cap=None):
        super(CinderObjectSerializer, self).__init__()
        self.version_cap = version_cap

        # NOTE(geguileo): During upgrades we will use a manifest to ensure that
        # all objects are properly backported.  This allows us to properly
        # backport child objects to the right version even if parent version
        # has not been bumped.
        if not version_cap or version_cap == OBJ_VERSIONS.get_current():
            self.manifest = None
        else:
            if version_cap not in OBJ_VERSIONS:
                raise exception.CappedVersionUnknown(version=version_cap)
            self.manifest = OBJ_VERSIONS[version_cap]

    def _get_capped_obj_version(self, obj):
        objname = obj.obj_name()
        version_dict = OBJ_VERSIONS.get(self.version_cap, {})
        version_cap = version_dict.get(objname, None)

        if version_cap:
            cap_tuple = versionutils.convert_version_to_tuple(version_cap)
            obj_tuple = versionutils.convert_version_to_tuple(obj.VERSION)
            if cap_tuple > obj_tuple:
                # NOTE(dulek): Do not set version cap to be higher than actual
                # object version as we don't support "forwardporting" of
                # objects. If service will receive an object that's too old it
                # should handle it explicitly.
                version_cap = None

        return version_cap

    def serialize_entity(self, context, entity):
        if isinstance(entity, (tuple, list, set, dict)):
            entity = self._process_iterable(context, self.serialize_entity,
                                            entity)
        elif (hasattr(entity, 'obj_to_primitive') and
              isinstance(entity.obj_to_primitive, abc.Callable)):
            # NOTE(dulek): Backport outgoing object to the capped version.
            backport_ver = self._get_capped_obj_version(entity)
            entity = entity.obj_to_primitive(backport_ver, self.manifest)
        return entity


def _log_backport(ovo, target_version):
    """Log backported versioned objects."""
    if target_version and target_version != ovo.VERSION:
        LOG.debug('Backporting %(obj_name)s from version %(src_vers)s '
                  'to version %(dst_vers)s',
                  {'obj_name': ovo.obj_name(),
                   'src_vers': ovo.VERSION,
                   'dst_vers': target_version})



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from http import HTTPStatus
from unittest import mock

import ddt
import fixtures
from oslo_concurrency import lockutils
from oslo_config import fixture as config_fixture
import oslo_messaging as messaging
from oslo_serialization import jsonutils
from oslo_utils import timeutils
import webob

from cinder.api.contrib import admin_actions
from cinder.api import microversions as mv
from cinder.backup import api as backup_api
from cinder.backup import rpcapi as backup_rpcapi
from cinder.common import constants
from cinder import context
from cinder import db
from cinder import exception
from cinder import objects
from cinder.objects import base as obj_base
from cinder.objects import fields
from cinder.scheduler import rpcapi as scheduler_rpcapi
from cinder.tests.unit.api import fakes
from cinder.tests.unit.api.v3 import fakes as v3_fakes
from cinder.tests.unit import cast_as_call
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import fake_snapshot
from cinder.tests.unit import test
from cinder.tests.unit import utils as test_utils
from cinder.volume import api as volume_api
from cinder.volume import rpcapi
from cinder.volume import volume_types


def app():
    # no auth, just let environ['cinder.context'] pass through
    api = fakes.router_v3.APIRouter()
    mapper = fakes.urlmap.URLMap()
    mapper['/v3'] = api
    return mapper


class BaseAdminTest(test.TestCase):
    def setUp(self):
        super(BaseAdminTest, self).setUp()
        self.volume_api = volume_api.API()
        # admin context
        self.ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID, True)

    def _create_volume(self, context, updates=None):
        db_volume = {'status': 'available',
                     'host': 'test',
                     'binary': constants.VOLUME_BINARY,
                     'availability_zone': 'fake_zone',
                     'attach_status': fields.VolumeAttachStatus.DETACHED,
                     'volume_type_id':
                         volume_types.get_default_volume_type()['id']}
        if updates:
            db_volume.update(updates)

        volume = objects.Volume(context=context, **db_volume)
        volume.create()
        return volume


@ddt.ddt
class AdminActionsTest(BaseAdminTest):
    def setUp(self):
        super(AdminActionsTest, self).setUp()

        self.tempdir = self.useFixture(fixtures.TempDir()).path
        self.fixture = self.useFixture(config_fixture.Config(lockutils.CONF))
        self.fixture.config(lock_path=self.tempdir,
                            group='oslo_concurrency')
        self.fixture.config(disable_process_locking=True,
                            group='oslo_concurrency')
        self.flags(transport_url='fake:/')

        cast_as_call.mock_cast_as_call(self.volume_api.volume_rpcapi.client)
        cast_as_call.mock_cast_as_call(self.volume_api.scheduler_rpcapi.client)

        # start service to handle rpc messages for attach requests
        self.svc = self.start_service('volume', host='test')
        self.patch(
            'cinder.objects.Service.get_minimum_obj_version',
            return_value=obj_base.OBJ_VERSIONS.get_current())

        def _get_minimum_rpc_version_mock(ctxt, binary):
            binary_map = {
                constants.VOLUME_BINARY: rpcapi.VolumeAPI,
                constants.BACKUP_BINARY: backup_rpcapi.BackupAPI,
                constants.SCHEDULER_BINARY: scheduler_rpcapi.SchedulerAPI,
            }
            return binary_map[binary].RPC_API_VERSION

        self.patch('cinder.objects.Service.get_minimum_rpc_version',
                   side_effect=_get_minimum_rpc_version_mock)
        self.controller = admin_actions.VolumeAdminController()

    def tearDown(self):
        self.svc.stop()
        super(AdminActionsTest, self).tearDown()

    def _issue_resource_reset(self, ctx, name, id, status):
        req = webob.Request.blank('/v3/%s/%s/%s/action' % (
            fake.PROJECT_ID, name, id))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes({'os-reset_status': status})
        req.environ['cinder.context'] = ctx
        resp = req.get_response(app())
        return resp

    def _issue_volume_reset(self, ctx, volume, updated_status):
        return self._issue_resource_reset(ctx,
                                          'volumes',
                                          volume['id'],
                                          updated_status)

    def _issue_snapshot_reset(self, ctx, snapshot, updated_status):
        return self._issue_resource_reset(ctx,
                                          'snapshots',
                                          snapshot.id,
                                          updated_status)

    def _issue_backup_reset(self, ctx, backup, updated_status):
        self.mock_object(backup_api.API,
                         '_get_available_backup_service_host',
                         return_value='testhost')
        return self._issue_resource_reset(ctx,
                                          'backups',
                                          backup['id'],
                                          updated_status)

    @ddt.data({'os-reset_status': {'status': 'creating'}},
              {'os-reset_status': {'status': 'available'}},
              {'os-reset_status': {'status': 'deleting'}},
              {'os-reset_status': {'status': 'error'}},
              {'os-reset_status': {'status': 'error_deleting'}},
              {'os-reset_status': {'attach_status':
                                   fields.VolumeAttachStatus.DETACHED}},
              {'os-reset_status': {'attach_status':
                                   fields.VolumeAttachStatus.ATTACHED}},
              {'os-reset_status': {'migration_status': 'migrating'}},
              {'os-reset_status': {'migration_status': 'completing'}},
              {'os-reset_status': {'migration_status': 'error'}},
              {'os-reset_status': {'migration_status': 'none'}},
              {'os-reset_status': {'migration_status': 'starting'}})
    def test_valid_updates(self, body):
        req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
            fake.PROJECT_ID, id))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        req.environ['cinder.context'] = self.ctx
        req.api_version_request = mv.get_api_version(mv.BASE_VERSION)
        vac = self.controller
        vac.validate_update(req, body=body)

    @ddt.data({'os-reset_status': {'status': None}},
              {'os-reset_status': {'attach_status': None}},
              {'os-reset_status': {'migration_status': None}},
              {'os-reset_status': {'status': "", 'attach_status': "",
                                   "migration_status": ""}})
    def test_invalid_updates(self, body):
        req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
            fake.PROJECT_ID, id))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        req.environ['cinder.context'] = self.ctx
        req.api_version_request = mv.get_api_version(mv.BASE_VERSION)
        vac = self.controller
        self.assertRaises(exception.InvalidParameterValue, vac.validate_update,
                          req, body=body)

    def test_reset_attach_status(self):
        volume = db.volume_create(self.ctx,
                                  {'attach_status':
                                   fields.VolumeAttachStatus.DETACHED,
                                   'volume_type_id': fake.VOLUME_TYPE_ID})

        resp = self._issue_volume_reset(self.ctx,
                                        volume,
                                        {'attach_status':
                                         fields.VolumeAttachStatus.ATTACHED})

        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)
        volume = db.volume_get(self.ctx, volume['id'])
        self.assertEqual(fields.VolumeAttachStatus.ATTACHED,
                         volume['attach_status'])

    def test_reset_attach_invalid_status(self):
        volume = db.volume_create(self.ctx,
                                  {'attach_status':
                                   fields.VolumeAttachStatus.DETACHED,
                                   'volume_type_id': fake.VOLUME_TYPE_ID})

        resp = self._issue_volume_reset(self.ctx,
                                        volume,
                                        {'attach_status': 'bogus-status'})

        self.assertEqual(HTTPStatus.BAD_REQUEST, resp.status_int)
        volume = db.volume_get(self.ctx, volume['id'])
        self.assertEqual(fields.VolumeAttachStatus.DETACHED,
                         volume['attach_status'])

    def test_reset_migration_invalid_status(self):
        volume = db.volume_create(self.ctx, {'migration_status': None,
                                             'volume_type_id':
                                                 fake.VOLUME_TYPE_ID})

        resp = self._issue_volume_reset(self.ctx,
                                        volume,
                                        {'migration_status': 'bogus-status'})

        self.assertEqual(HTTPStatus.BAD_REQUEST, resp.status_int)
        volume = db.volume_get(self.ctx, volume['id'])
        self.assertIsNone(volume['migration_status'])

    def test_reset_migration_status(self):
        volume = db.volume_create(self.ctx, {'migration_status': None,
                                             'volume_type_id':
                                                 fake.VOLUME_TYPE_ID})

        resp = self._issue_volume_reset(self.ctx,
                                        volume,
                                        {'migration_status': 'migrating'})

        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)
        volume = db.volume_get(self.ctx, volume['id'])
        self.assertEqual('migrating', volume['migration_status'])

    def test_reset_status_as_admin(self):
        volume = db.volume_create(self.ctx, {'status': 'available',
                                             'volume_type_id':
                                                 fake.VOLUME_TYPE_ID})

        resp = self._issue_volume_reset(self.ctx,
                                        volume,
                                        {'status': 'error'})

        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)
        volume = db.volume_get(self.ctx, volume['id'])
        self.assertEqual('error', volume['status'])

    @mock.patch('cinder.objects.volume.Volume.get_by_id')
    def test_reset_status_as_non_admin(self, fake_get):
        ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID)
        volume = db.volume_create(self.ctx,
                                  {'status': 'error', 'size': 1,
                                   'volume_type_id': fake.VOLUME_TYPE_ID})
        fake_get.return_value = volume
        resp = self._issue_volume_reset(ctx,
                                        volume,
                                        {'status': 'error'})

        # request is not authorized
        self.assertEqual(HTTPStatus.FORBIDDEN, resp.status_int)
        volume = db.volume_get(self.ctx, volume['id'])
        # status is still 'error'
        self.assertEqual('error', volume['status'])

    def test_backup_reset_status_as_admin(self):
        volume = db.volume_create(self.ctx, {'status': 'available',
                                             'user_id': fake.USER_ID,
                                             'project_id': fake.PROJECT_ID,
                                             'volume_type_id':
                                                 fake.VOLUME_TYPE_ID})
        backup = db.backup_create(self.ctx,
                                  {'status': fields.BackupStatus.AVAILABLE,
                                   'size': 1,
                                   'volume_id': volume['id'],
                                   'user_id': fake.USER_ID,
                                   'project_id': fake.PROJECT_ID,
                                   'host': 'test'})

        resp = self._issue_backup_reset(self.ctx,
                                        backup,
                                        {'status': fields.BackupStatus.ERROR})

        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)

    def test_backup_reset_status_as_non_admin(self):
        ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID)
        backup = test_utils.create_backup(ctx, status='available')
        resp = self._issue_backup_reset(ctx,
                                        backup,
                                        {'status': fields.BackupStatus.ERROR})
        # request is not authorized
        self.assertEqual(HTTPStatus.FORBIDDEN, resp.status_int)

    def test_backup_reset_status(self):
        volume = db.volume_create(self.ctx,
                                  {'status': 'available', 'host': 'test',
                                   'provider_location': '', 'size': 1,
                                   'volume_type_id': fake.VOLUME_TYPE_ID})
        backup = db.backup_create(self.ctx,
                                  {'status': fields.BackupStatus.AVAILABLE,
                                   'volume_id': volume['id'],
                                   'user_id': fake.USER_ID,
                                   'project_id': fake.PROJECT_ID,
                                   'host': 'test'})

        resp = self._issue_backup_reset(self.ctx,
                                        backup,
                                        {'status': fields.BackupStatus.ERROR})

        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)

    @ddt.data({'status': None}, {'status': 'restoring'})
    def test_invalid_status_for_backup(self, status):
        volume = db.volume_create(self.ctx,
                                  {'status': 'available', 'host': 'test',
                                   'provider_location': '', 'size': 1,
                                   'volume_type_id': fake.VOLUME_TYPE_ID})
        backup = db.backup_create(self.ctx, {'status': 'available',
                                             'volume_id': volume['id']})
        resp = self._issue_backup_reset(self.ctx,
                                        backup, status)
        self.assertEqual(HTTPStatus.BAD_REQUEST, resp.status_int)

    def test_backup_reset_status_with_invalid_backup(self):
        volume = db.volume_create(self.ctx,
                                  {'status': 'available', 'host': 'test',
                                   'provider_location': '', 'size': 1,
                                   'volume_type_id': fake.VOLUME_TYPE_ID})
        backup = db.backup_create(self.ctx,
                                  {'status': fields.BackupStatus.AVAILABLE,
                                   'volume_id': volume['id'],
                                   'user_id': fake.USER_ID,
                                   'project_id': fake.PROJECT_ID})

        backup['id'] = fake.BACKUP_ID
        resp = self._issue_backup_reset(self.ctx,
                                        backup,
                                        {'status': fields.BackupStatus.ERROR})

        # Should raise 404 if backup doesn't exist.
        self.assertEqual(HTTPStatus.NOT_FOUND, resp.status_int)

    @ddt.data({'os-reset_status': {}})
    def test_backup_reset_status_with_invalid_body(self, body):
        volume = db.volume_create(self.ctx,
                                  {'status': 'available', 'host': 'test',
                                   'provider_location': '', 'size': 1,
                                   'volume_type_id': fake.VOLUME_TYPE_ID})
        backup = db.backup_create(self.ctx,
                                  {'status': fields.BackupStatus.AVAILABLE,
                                   'volume_id': volume['id'],
                                   'user_id': fake.USER_ID,
                                   'project_id': fake.PROJECT_ID})

        req = webob.Request.blank('/v3/%s/%s/%s/action' % (
            fake.PROJECT_ID, 'backups', backup['id']))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        req.environ['cinder.context'] = self.ctx
        resp = req.get_response(app())
        self.assertEqual(HTTPStatus.BAD_REQUEST, resp.status_int)

    def test_malformed_reset_status_body(self):
        volume = db.volume_create(self.ctx, {'status': 'available', 'size': 1,
                                             'volume_type_id':
                                                 fake.VOLUME_TYPE_ID})

        resp = self._issue_volume_reset(self.ctx,
                                        volume,
                                        {'x-status': 'bad'})

        self.assertEqual(HTTPStatus.BAD_REQUEST, resp.status_int)
        volume = db.volume_get(self.ctx, volume['id'])
        self.assertEqual('available', volume['status'])

    def test_invalid_status_for_volume(self):
        volume = db.volume_create(self.ctx, {'status': 'available', 'size': 1,
                                             'volume_type_id':
                                                 fake.VOLUME_TYPE_ID})
        resp = self._issue_volume_reset(self.ctx,
                                        volume,
                                        {'status': 'invalid'})

        self.assertEqual(HTTPStatus.BAD_REQUEST, resp.status_int)
        volume = db.volume_get(self.ctx, volume['id'])
        self.assertEqual('available', volume['status'])

    def test_reset_status_for_missing_volume(self):
        req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
            fake.PROJECT_ID, fake.WILL_NOT_BE_FOUND_ID))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        body = {'os-reset_status': {'status': 'available'}}
        req.body = jsonutils.dump_as_bytes(body)
        req.environ['cinder.context'] = self.ctx
        resp = req.get_response(app())
        self.assertEqual(HTTPStatus.NOT_FOUND, resp.status_int)
        self.assertRaises(exception.NotFound, db.volume_get, self.ctx,
                          fake.WILL_NOT_BE_FOUND_ID)

    def test_reset_attached_status(self):
        # current status is available
        volume = self._create_volume(self.ctx, {'provider_location': '',
                                                'size': 1})
        self.volume_api.reserve_volume(self.ctx, volume)
        mountpoint = '/dev/vdb'
        attachment = self.volume_api.attach(self.ctx, volume, fake.INSTANCE_ID,
                                            None, mountpoint, 'rw')
        # volume is attached
        volume = db.volume_get(self.ctx.elevated(), volume['id'])
        attachment = db.volume_attachment_get(self.ctx, attachment['id'])

        self.assertEqual('in-use', volume['status'])
        self.assertEqual(fields.VolumeAttachStatus.ATTACHED,
                         volume['attach_status'])
        self.assertEqual(fake.INSTANCE_ID, attachment['instance_uuid'])
        self.assertEqual(mountpoint, attachment['mountpoint'])
        self.assertEqual('attached', attachment['attach_status'])
        admin_metadata = volume['volume_admin_metadata']
        self.assertEqual(2, len(admin_metadata))
        self.assertEqual('readonly', admin_metadata[0]['key'])
        self.assertEqual('False', admin_metadata[0]['value'])
        self.assertEqual('attached_mode', admin_metadata[1]['key'])
        self.assertEqual('rw', admin_metadata[1]['value'])

        # Reset attach_status
        resp = self._issue_volume_reset(
            self.ctx,
            volume,
            {'status': 'available',
             'attach_status': fields.VolumeAttachStatus.DETACHED})
        # request is accepted
        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)

        # volume is detached
        volume = db.volume_get(self.ctx, volume['id'])
        self.assertEqual('detached', volume['attach_status'])
        self.assertEqual('available', volume['status'])
        admin_metadata = volume['volume_admin_metadata']
        self.assertEqual(1, len(admin_metadata))
        self.assertEqual('readonly', admin_metadata[0]['key'])
        self.assertEqual('False', admin_metadata[0]['value'])
        self.assertRaises(exception.VolumeAttachmentNotFound,
                          db.volume_attachment_get,
                          self.ctx, attachment['id'])

    def test_invalid_reset_attached_status(self):
        volume = db.volume_create(self.ctx,
                                  {'status': 'available', 'host': 'test',
                                   'provider_location': '', 'size': 1,
                                   'attach_status':
                                       fields.VolumeAttachStatus.DETACHED,
                                   'volume_type_id': fake.VOLUME_TYPE_ID})
        resp = self._issue_volume_reset(
            self.ctx,
            volume,
            {'status': 'available',
             'attach_status': fields.VolumeAttachStatus.ERROR_DETACHING})
        self.assertEqual(HTTPStatus.BAD_REQUEST, resp.status_int)
        volume = db.volume_get(self.ctx, volume['id'])
        self.assertEqual('available', volume['status'])
        self.assertEqual(fields.VolumeAttachStatus.DETACHED,
                         volume['attach_status'])

    def test_snapshot_reset_status(self):
        volume = db.volume_create(self.ctx,
                                  {'status': 'available', 'host': 'test',
                                   'provider_location': '', 'size': 1,
                                   'availability_zone': 'test',
                                   'attach_status':
                                       fields.VolumeAttachStatus.DETACHED,
                                   'volume_type_id': fake.VOLUME_TYPE_ID})
        kwargs = {
            'volume_id': volume['id'],
            'cgsnapshot_id': None,
            'user_id': self.ctx.user_id,
            'project_id': self.ctx.project_id,
            'status': fields.SnapshotStatus.ERROR_DELETING,
            'progress': '0%',
            'volume_size': volume['size'],
            'metadata': {}
        }
        snapshot = objects.Snapshot(context=self.ctx, **kwargs)
        snapshot.create()
        self.addCleanup(snapshot.destroy)

        resp = self._issue_snapshot_reset(self.ctx, snapshot,
                                          {'status':
                                           fields.SnapshotStatus.ERROR})

        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)
        snapshot = objects.Snapshot.get_by_id(self.ctx, snapshot['id'])
        self.assertEqual(fields.SnapshotStatus.ERROR, snapshot.status)

    @ddt.data({'status': None}, {'status': 'attaching'})
    def test_invalid_status_for_snapshot(self, updated_status):
        volume = db.volume_create(self.ctx,
                                  {'status': 'available', 'host': 'test',
                                   'provider_location': '', 'size': 1,
                                   'volume_type_id': fake.VOLUME_TYPE_ID})
        snapshot = objects.Snapshot(self.ctx,
                                    status=fields.SnapshotStatus.AVAILABLE,
                                    volume_id=volume['id'])
        snapshot.create()
        self.addCleanup(snapshot.destroy)

        resp = self._issue_snapshot_reset(self.ctx, snapshot, updated_status)

        self.assertEqual(HTTPStatus.BAD_REQUEST, resp.status_int)
        self.assertEqual(fields.SnapshotStatus.AVAILABLE, snapshot.status)

    @ddt.data({'os-reset_status': {}})
    def test_snapshot_reset_status_with_invalid_body(self, body):
        volume = db.volume_create(self.ctx,
                                  {'status': 'available', 'host': 'test',
                                   'provider_location': '', 'size': 1,
                                   'volume_type_id': fake.VOLUME_TYPE_ID})
        snapshot = objects.Snapshot(self.ctx,
                                    status=fields.SnapshotStatus.AVAILABLE,
                                    volume_id=volume['id'],
                                    volume_tpe_id=volume['volume_type_id'])
        snapshot.create()
        self.addCleanup(snapshot.destroy)

        req = webob.Request.blank('/v3/%s/%s/%s/action' % (
            fake.PROJECT_ID, 'snapshots', snapshot['id']))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes(body)
        req.environ['cinder.context'] = self.ctx
        resp = req.get_response(app())
        self.assertEqual(HTTPStatus.BAD_REQUEST, resp.status_int)

    def test_force_delete(self):
        # current status is creating
        volume = self._create_volume(self.ctx, {'size': 1, 'host': None})
        req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
            fake.PROJECT_ID, volume['id']))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes({'os-force_delete': {}})
        # attach admin context to request
        req.environ['cinder.context'] = self.ctx
        resp = req.get_response(app())
        # request is accepted
        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)
        # volume is deleted
        self.assertRaises(exception.NotFound, objects.Volume.get_by_id,
                          self.ctx, volume.id)

    @mock.patch.object(volume_api.API, 'delete_snapshot', return_value=True)
    @mock.patch('cinder.objects.Snapshot.get_by_id')
    @mock.patch.object(db, 'snapshot_get')
    @mock.patch.object(db, 'volume_get')
    def test_force_delete_snapshot(self, volume_get, snapshot_get, get_by_id,
                                   delete_snapshot):
        volume = v3_fakes.create_volume(fake.VOLUME_ID)
        snapshot = v3_fakes.fake_snapshot(fake.SNAPSHOT_ID)
        snapshot_obj = fake_snapshot.fake_snapshot_obj(self.ctx, **snapshot)
        volume_get.return_value = volume
        snapshot_get.return_value = snapshot
        get_by_id.return_value = snapshot_obj

        path = '/v3/%s/snapshots/%s/action' % (
            fake.PROJECT_ID, snapshot['id'])
        req = webob.Request.blank(path)
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes({'os-force_delete': {}})
        # attach admin context to request
        req.environ['cinder.context'] = self.ctx
        resp = req.get_response(app())
        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)

    def _migrate_volume_prep(self):
        # create volume's current host and the destination host
        db.service_create(self.ctx,
                          {'host': 'test',
                           'topic': constants.VOLUME_TOPIC,
                           'binary': constants.VOLUME_BINARY,
                           'created_at': timeutils.utcnow()})
        db.service_create(self.ctx,
                          {'host': 'test2',
                           'topic': constants.VOLUME_TOPIC,
                           'binary': constants.VOLUME_BINARY,
                           'created_at': timeutils.utcnow()})
        db.service_create(self.ctx,
                          {'host': 'clustered_host',
                           'topic': constants.VOLUME_TOPIC,
                           'binary': constants.VOLUME_BINARY,
                           'cluster_name': 'cluster',
                           'created_at': timeutils.utcnow()})
        db.cluster_create(self.ctx,
                          {'name': 'cluster',
                           'binary': constants.VOLUME_BINARY})
        # current status is available
        volume = self._create_volume(self.ctx)
        return volume

    def _migrate_volume_3_exec(self, ctx, volume, host, expected_status,
                               force_host_copy=False, version=None,
                               cluster=None):
        # build request to migrate to host
        req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
            fake.PROJECT_ID, volume['id']))
        req.method = 'POST'
        body = {'os-migrate_volume': {'host': host,
                                      'force_host_copy': force_host_copy}}
        version = version or mv.BASE_VERSION
        req.headers = mv.get_mv_header(version)
        req.headers['Content-Type'] = 'application/json'
        req.api_version_request = mv.get_api_version(version)
        if version == mv.VOLUME_MIGRATE_CLUSTER:
            body['os-migrate_volume']['cluster'] = cluster
        req.body = jsonutils.dump_as_bytes(body)
        req.environ['cinder.context'] = ctx
        resp = req.get_response(app())

        # verify status
        self.assertEqual(expected_status, resp.status_int)
        volume = db.volume_get(self.ctx, volume['id'])
        return volume

    @ddt.data(mv.BASE_VERSION,
              mv.get_prior_version(mv.VOLUME_MIGRATE_CLUSTER),
              mv.VOLUME_MIGRATE_CLUSTER)
    def test_migrate_volume_success_3(self, version):
        expected_status = HTTPStatus.ACCEPTED
        host = 'test2'
        volume = self._migrate_volume_prep()
        volume = self._migrate_volume_3_exec(self.ctx, volume, host,
                                             expected_status, version=version)
        self.assertEqual('starting', volume['migration_status'])

    def test_migrate_volume_success_cluster(self):
        expected_status = HTTPStatus.ACCEPTED
        # We cannot provide host and cluster, so send host to None
        host = None
        cluster = 'cluster'
        volume = self._migrate_volume_prep()
        volume = self._migrate_volume_3_exec(self.ctx, volume, host,
                                             expected_status,
                                             version=mv.VOLUME_MIGRATE_CLUSTER,
                                             cluster=cluster)
        self.assertEqual('starting', volume['migration_status'])

    def test_migrate_volume_fail_host_and_cluster(self):
        # We cannot send host and cluster in the request
        host = 'test2'
        cluster = 'cluster'
        volume = self._migrate_volume_prep()
        expected_status = HTTPStatus.BAD_REQUEST
        self._migrate_volume_3_exec(self.ctx, volume, host, expected_status,
                                    version=mv.VOLUME_MIGRATE_CLUSTER,
                                    cluster=cluster)

    def _migrate_volume_exec(self, ctx, volume, host, expected_status,
                             force_host_copy=False, lock_volume=False):
        # build request to migrate to host
        req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
            fake.PROJECT_ID, volume['id']))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        body = {'os-migrate_volume': {'host': host,
                                      'force_host_copy': force_host_copy,
                                      'lock_volume': lock_volume}}
        req.body = jsonutils.dump_as_bytes(body)
        req.environ['cinder.context'] = ctx
        resp = req.get_response(app())
        # verify status
        self.assertEqual(expected_status, resp.status_int)
        volume = db.volume_get(self.ctx, volume['id'])
        return volume

    def test_migrate_volume_success(self):
        expected_status = HTTPStatus.ACCEPTED
        host = 'test2'
        volume = self._migrate_volume_prep()
        volume = self._migrate_volume_exec(self.ctx, volume, host,
                                           expected_status)
        self.assertEqual('starting', volume['migration_status'])

    def test_migrate_volume_fail_replication(self):
        expected_status = HTTPStatus.BAD_REQUEST
        host = 'test2'
        volume = self._migrate_volume_prep()
        # current status is available
        volume = self._create_volume(self.ctx,
                                     {'provider_location': '',
                                      'attach_status': None,
                                      'replication_status': 'active'})
        volume = self._migrate_volume_exec(self.ctx, volume, host,
                                           expected_status)

    def test_migrate_volume_replication_not_caple_success(self):
        expected_status = HTTPStatus.ACCEPTED
        host = 'test2'
        volume = self._migrate_volume_prep()
        # current status is available
        volume = self._create_volume(self.ctx,
                                     {'provider_location': '',
                                      'attach_status': None,
                                      'replication_status': 'not-capable'})
        volume = self._migrate_volume_exec(self.ctx, volume, host,
                                           expected_status)

    @mock.patch("cinder.volume.api.API.get")
    def test_migrate_volume_as_non_admin(self, fake_get):
        expected_status = HTTPStatus.FORBIDDEN
        host = 'test2'
        ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID)
        volume = self._migrate_volume_prep()
        fake_get.return_value = volume
        self._migrate_volume_exec(ctx, volume, host, expected_status)

    def test_migrate_volume_without_host_parameter(self):
        expected_status = HTTPStatus.BAD_REQUEST
        host = 'test3'
        volume = self._migrate_volume_prep()
        # build request to migrate without host
        req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
            fake.PROJECT_ID, volume['id']))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        body = {'os-migrate_volume': {'host': host,
                                      'force_host_copy': False}}
        req.body = jsonutils.dump_as_bytes(body)
        req.environ['cinder.context'] = self.ctx
        resp = req.get_response(app())
        # verify status
        self.assertEqual(expected_status, resp.status_int)

    def test_migrate_volume_host_no_exist(self):
        expected_status = HTTPStatus.BAD_REQUEST
        host = 'test3'
        volume = self._migrate_volume_prep()
        self._migrate_volume_exec(self.ctx, volume, host, expected_status)

    def test_migrate_volume_same_host(self):
        expected_status = HTTPStatus.BAD_REQUEST
        host = 'test'
        volume = self._migrate_volume_prep()
        self._migrate_volume_exec(self.ctx, volume, host, expected_status)

    def test_migrate_volume_migrating(self):
        expected_status = HTTPStatus.BAD_REQUEST
        host = 'test2'
        volume = self._migrate_volume_prep()
        volume.migration_status = 'migrating'
        volume.save()
        self._migrate_volume_exec(self.ctx, volume, host, expected_status)

    def test_migrate_volume_with_snap(self):
        expected_status = HTTPStatus.BAD_REQUEST
        host = 'test2'
        volume = self._migrate_volume_prep()
        snap = objects.Snapshot(self.ctx, volume_id=volume['id'])
        snap.create()
        self.addCleanup(snap.destroy)
        self._migrate_volume_exec(self.ctx, volume, host, expected_status)

    @ddt.data('force_host_copy', None, '  true  ', 0)
    def test_migrate_volume_bad_force_host_copy(self, force_host_copy):
        expected_status = HTTPStatus.BAD_REQUEST
        host = 'test2'
        volume = self._migrate_volume_prep()
        self._migrate_volume_exec(self.ctx, volume, host, expected_status,
                                  force_host_copy=force_host_copy)

    @ddt.data('lock_volume', None, '  true  ', 0)
    def test_migrate_volume_bad_lock_volume(self, lock_volume):
        expected_status = HTTPStatus.BAD_REQUEST
        host = 'test2'
        volume = self._migrate_volume_prep()
        self._migrate_volume_exec(self.ctx, volume, host, expected_status,
                                  lock_volume=lock_volume)

    @ddt.data('true', False, '1', '0')
    def test_migrate_volume_valid_lock_volume(self, lock_volume):
        expected_status = HTTPStatus.ACCEPTED
        host = 'test2'
        volume = self._migrate_volume_prep()
        self._migrate_volume_exec(self.ctx, volume, host, expected_status,
                                  lock_volume=lock_volume)

    def _migrate_volume_comp_exec(self, ctx, volume, new_volume, error,
                                  expected_status, expected_id, no_body=False):
        req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
            fake.PROJECT_ID, volume['id']))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        body = {'new_volume': new_volume['id'], 'error': error}
        if no_body:
            body = {'': body}
        else:
            body = {'os-migrate_volume_completion': body}
        req.body = jsonutils.dump_as_bytes(body)
        req.environ['cinder.context'] = ctx
        resp = req.get_response(app())
        resp_dict = resp.json
        # verify status
        self.assertEqual(expected_status, resp.status_int)
        if expected_id:
            self.assertEqual(expected_id, resp_dict['save_volume_id'])
        else:
            self.assertNotIn('save_volume_id', resp_dict)

    @mock.patch("cinder.volume.api.API.get")
    def test_migrate_volume_comp_as_non_admin(self, fake_get):
        volume = db.volume_create(self.ctx, {'id': fake.VOLUME_ID,
                                             'volume_type_id':
                                                 fake.VOLUME_TYPE_ID})
        new_volume = db.volume_create(self.ctx, {'id': fake.VOLUME2_ID,
                                                 'volume_type_id':
                                                     fake.VOLUME_TYPE_ID})
        expected_status = HTTPStatus.FORBIDDEN
        expected_id = None
        fake_get.return_value = volume
        ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID)
        self._migrate_volume_comp_exec(ctx, volume, new_volume, False,
                                       expected_status, expected_id)

    def test_migrate_volume_comp_no_mig_status(self):
        volume1 = self._create_volume(self.ctx, {'migration_status': 'foo'})
        volume2 = self._create_volume(self.ctx, {'migration_status': None})

        expected_status = HTTPStatus.BAD_REQUEST
        expected_id = None
        self._migrate_volume_comp_exec(self.ctx, volume1, volume2, False,
                                       expected_status, expected_id)
        self._migrate_volume_comp_exec(self.ctx, volume2, volume1, False,
                                       expected_status, expected_id)

    def test_migrate_volume_comp_bad_mig_status(self):
        volume1 = self._create_volume(self.ctx,
                                      {'migration_status': 'migrating'})
        volume2 = self._create_volume(self.ctx,
                                      {'migration_status': 'target:foo'})
        expected_status = HTTPStatus.BAD_REQUEST
        expected_id = None
        self._migrate_volume_comp_exec(self.ctx, volume1, volume2, False,
                                       expected_status, expected_id)

    def test_migrate_volume_comp_no_action(self):
        volume = db.volume_create(self.ctx, {'id': fake.VOLUME_ID,
                                             'volume_type_id':
                                                 fake.VOLUME_TYPE_ID})
        new_volume = db.volume_create(self.ctx, {'id': fake.VOLUME2_ID,
                                                 'volume_type_id':
                                                     fake.VOLUME_TYPE_ID})
        expected_status = HTTPStatus.BAD_REQUEST
        expected_id = None
        ctx = context.RequestContext(fake.USER_ID, fake.PROJECT_ID)
        self._migrate_volume_comp_exec(ctx, volume, new_volume, False,
                                       expected_status, expected_id, True)

    def test_migrate_volume_comp_from_nova(self):
        volume = self._create_volume(self.ctx, {'status': 'in-use',
                                                'migration_status': None,
                                                'attach_status':
                                                    fields.VolumeAttachStatus.
                                                    ATTACHED})
        new_volume = self._create_volume(self.ctx,
                                         {'migration_status': None,
                                          'attach_status':
                                              fields.VolumeAttachStatus.
                                              DETACHED})
        expected_status = HTTPStatus.OK
        expected_id = new_volume.id
        self._migrate_volume_comp_exec(self.ctx, volume, new_volume, False,
                                       expected_status, expected_id)

    def test_migrate_volume_comp_no_new_volume(self):
        volume = db.volume_create(self.ctx, {'id': fake.VOLUME_ID,
                                             'volume_type_id':
                                                 fake.VOLUME_TYPE_ID})
        req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
            fake.PROJECT_ID, volume['id']))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        body = {'os-migrate_volume_completion': {'error': False}}
        req.body = jsonutils.dump_as_bytes(body)
        req.environ['cinder.context'] = self.ctx
        resp = req.get_response(app())
        res_dict = jsonutils.loads(resp.body)
        self.assertEqual(HTTPStatus.BAD_REQUEST,
                         res_dict['badRequest']['code'])

    @mock.patch('cinder.backup.rpcapi.BackupAPI.delete_backup', mock.Mock())
    @mock.patch('cinder.db.service_get_all')
    @mock.patch('cinder.backup.api.API._check_support_to_force_delete')
    def _force_delete_backup_util(self, test_status, mock_check_support,
                                  mock_service_get_all):
        mock_service_get_all.return_value = [
            {'availability_zone': "az1", 'host': 'testhost',
             'disabled': 0, 'updated_at': timeutils.utcnow(),
             'uuid': 'a3a593da-7f8d-4bb7-8b4c-f2bc1e0b4824'}]
        # admin context
        mock_check_support.return_value = True
        # current status is dependent on argument: test_status.
        backup = test_utils.create_backup(self.ctx, status=test_status,
                                          size=1, availability_zone='az1',
                                          host='testhost')
        req = webob.Request.blank('/v3/%s/backups/%s/action' % (
            fake.PROJECT_ID, backup.id))
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes({'os-force_delete': {}})
        req.environ['cinder.context'] = self.ctx
        res = req.get_response(app())

        backup.refresh()
        self.assertEqual(HTTPStatus.ACCEPTED, res.status_int)
        self.assertEqual('deleting', backup.status)
        backup.destroy()

    def test_delete_backup_force_when_creating(self):
        self._force_delete_backup_util('creating')

    def test_delete_backup_force_when_deleting(self):
        self._force_delete_backup_util('deleting')

    def test_delete_backup_force_when_restoring(self):
        self._force_delete_backup_util('restoring')

    def test_delete_backup_force_when_available(self):
        self._force_delete_backup_util('available')

    def test_delete_backup_force_when_error(self):
        self._force_delete_backup_util('error')

    def test_delete_backup_force_when_error_deleting(self):
        self._force_delete_backup_util('error_deleting')

    @mock.patch('cinder.backup.rpcapi.BackupAPI.check_support_to_force_delete',
                return_value=False)
    def test_delete_backup_force_when_not_supported(self, mock_check_support):
        # admin context
        self.override_config('backup_driver', 'cinder.backup.drivers.ceph')
        backup = test_utils.create_backup(self.ctx, size=1)
        req = webob.Request.blank('/v3/%s/backups/%s/action' % (
            fake.PROJECT_ID, backup.id))
        req.method = 'POST'
        req.headers['Content-Type'] = 'application/json'
        req.body = jsonutils.dump_as_bytes({'os-force_delete': {}})
        req.environ['cinder.context'] = self.ctx
        res = req.get_response(app())
        self.assertEqual(HTTPStatus.METHOD_NOT_ALLOWED, res.status_int)


class AdminActionsAttachDetachTest(BaseAdminTest):
    def setUp(self):
        super(AdminActionsAttachDetachTest, self).setUp()
        # start service to handle rpc messages for attach requests
        self.svc = self.start_service('volume', host='test')

    def tearDown(self):
        self.svc.stop()
        super(AdminActionsAttachDetachTest, self).tearDown()

    def test_force_detach_instance_attached_volume(self):
        # current status is available
        volume = self._create_volume(self.ctx, {'provider_location': '',
                                                'size': 1})
        connector = {'initiator': 'iqn.2012-07.org.fake:01'}

        self.volume_api.reserve_volume(self.ctx, volume)
        mountpoint = '/dev/vbd'
        attachment = self.volume_api.attach(self.ctx, volume, fake.INSTANCE_ID,
                                            None, mountpoint, 'rw')
        # volume is attached
        volume.refresh()
        self.assertEqual('in-use', volume.status)
        self.assertEqual(fake.INSTANCE_ID, attachment['instance_uuid'])
        self.assertEqual(mountpoint, attachment['mountpoint'])
        self.assertEqual(fields.VolumeAttachStatus.ATTACHED,
                         attachment['attach_status'])
        admin_metadata = volume.admin_metadata
        self.assertEqual(2, len(admin_metadata))
        self.assertEqual('False', admin_metadata['readonly'])
        self.assertEqual('rw', admin_metadata['attached_mode'])
        conn_info = self.volume_api.initialize_connection(self.ctx,
                                                          volume,
                                                          connector)
        self.assertEqual('rw', conn_info['data']['access_mode'])
        # build request to force detach
        req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
            fake.PROJECT_ID, volume.id))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        # request status of 'error'
        body = {'os-force_detach': {'attachment_id': attachment['id'],
                                    'connector': connector}}
        req.body = jsonutils.dump_as_bytes(body)
        # attach admin context to request
        req.environ['cinder.context'] = self.ctx
        # make request
        resp = req.get_response(app())
        # request is accepted
        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)
        volume.refresh()
        self.assertRaises(exception.VolumeAttachmentNotFound,
                          db.volume_attachment_get,
                          self.ctx, attachment['id'])

        # status changed to 'available'
        self.assertEqual('available', volume.status)
        admin_metadata = volume.admin_metadata
        self.assertEqual(1, len(admin_metadata))
        self.assertEqual('False', admin_metadata['readonly'])

    def test_force_detach_host_attached_volume(self):
        # current status is available
        volume = self._create_volume(self.ctx, {'provider_location': '',
                                                'size': 1})
        connector = {'initiator': 'iqn.2012-07.org.fake:01'}

        self.volume_api.initialize_connection(self.ctx, volume, connector)
        mountpoint = '/dev/vbd'
        host_name = 'fake-host'
        attachment = self.volume_api.attach(self.ctx, volume, None, host_name,
                                            mountpoint, 'ro')
        # volume is attached
        volume.refresh()
        self.assertEqual('in-use', volume.status)
        self.assertIsNone(attachment['instance_uuid'])
        self.assertEqual(host_name, attachment['attached_host'])
        self.assertEqual(mountpoint, attachment['mountpoint'])
        self.assertEqual(fields.VolumeAttachStatus.ATTACHED,
                         attachment['attach_status'])
        admin_metadata = volume.admin_metadata
        self.assertEqual(2, len(admin_metadata))
        self.assertEqual('False', admin_metadata['readonly'])
        self.assertEqual('ro', admin_metadata['attached_mode'])
        conn_info = self.volume_api.initialize_connection(self.ctx,
                                                          volume, connector)
        self.assertEqual('ro', conn_info['data']['access_mode'])
        # build request to force detach
        req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
            fake.PROJECT_ID, volume.id))
        req.method = 'POST'
        req.headers['content-type'] = 'application/json'
        # request status of 'error'
        body = {'os-force_detach': {'attachment_id': attachment['id'],
                                    'connector': connector}}
        req.body = jsonutils.dump_as_bytes(body)
        # attach admin context to request
        req.environ['cinder.context'] = self.ctx
        # make request
        resp = req.get_response(app())
        # request is accepted
        self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)
        volume.refresh()
        self.assertRaises(exception.VolumeAttachmentNotFound,
                          db.volume_attachment_get,
                          self.ctx, attachment['id'])
        # status changed to 'available'
        self.assertEqual('available', volume['status'])
        admin_metadata = volume['admin_metadata']
        self.assertEqual(1, len(admin_metadata))
        self.assertEqual('False', admin_metadata['readonly'])

    def test_volume_force_detach_raises_remote_error(self):
        # current status is available
        volume = self._create_volume(self.ctx, {'provider_location': '',
                                                'size': 1})
        connector = {'initiator': 'iqn.2012-07.org.fake:01'}

        self.volume_api.reserve_volume(self.ctx, volume)
        mountpoint = '/dev/vbd'
        attachment = self.volume_api.attach(self.ctx, volume, fake.INSTANCE_ID,
                                            None, mountpoint, 'rw')
        # volume is attached
        volume.refresh()
        self.assertEqual('in-use', volume.status)
        self.assertEqual(fake.INSTANCE_ID, attachment['instance_uuid'])
        self.assertEqual(mountpoint, attachment['mountpoint'])
        self.assertEqual(fields.VolumeAttachStatus.ATTACHED,
                         attachment['attach_status'])
        admin_metadata = volume.admin_metadata
        self.assertEqual(2, len(admin_metadata))
        self.assertEqual('False', admin_metadata['readonly'])
        self.assertEqual('rw', admin_metadata['attached_mode'])
        conn_info = self.volume_api.initialize_connection(self.ctx,
                                                          volume,
                                                          connector)
        self.assertEqual('rw', conn_info['data']['access_mode'])
        # build request to force detach
        volume_remote_error = \
            messaging.RemoteError(exc_type='VolumeAttachmentNotFound')
        with mock.patch.object(volume_api.API, 'detach',
                               side_effect=volume_remote_error):
            req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
                fake.PROJECT_ID, volume.id))
            req.method = 'POST'
            req.headers['content-type'] = 'application/json'
            body = {'os-force_detach': {'attachment_id': fake.ATTACHMENT_ID}}
            req.body = jsonutils.dump_as_bytes(body)
            # attach admin context to request
            req.environ['cinder.context'] = self.ctx
            # make request
            resp = req.get_response(app())
            self.assertEqual(HTTPStatus.BAD_REQUEST, resp.status_int)

        # test for VolumeBackendAPIException
        volume_remote_error = (
            messaging.RemoteError(exc_type='VolumeBackendAPIException'))
        with mock.patch.object(volume_api.API, 'detach',
                               side_effect=volume_remote_error):
            req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
                fake.PROJECT_ID, volume.id))
            req.method = 'POST'
            req.headers['content-type'] = 'application/json'
            body = {'os-force_detach': {'attachment_id': fake.ATTACHMENT_ID,
                                        'connector': connector}}
            req.body = jsonutils.dump_as_bytes(body)

            # attach admin context to request
            req.environ['cinder.context'] = self.ctx
            # make request
            self.assertRaises(messaging.RemoteError,
                              req.get_response,
                              app())

    def test_volume_force_detach_raises_db_error(self):
        # In case of DB error 500 error code is returned to user
        # current status is available
        volume = self._create_volume(self.ctx, {'provider_location': '',
                                                'size': 1})
        connector = {'initiator': 'iqn.2012-07.org.fake:01'}

        self.volume_api.reserve_volume(self.ctx, volume)
        mountpoint = '/dev/vbd'
        attachment = self.volume_api.attach(self.ctx, volume, fake.INSTANCE_ID,
                                            None, mountpoint, 'rw')
        # volume is attached
        volume.refresh()
        self.assertEqual('in-use', volume.status)
        self.assertEqual(fake.INSTANCE_ID, attachment['instance_uuid'])
        self.assertEqual(mountpoint, attachment['mountpoint'])
        self.assertEqual(fields.VolumeAttachStatus.ATTACHED,
                         attachment['attach_status'])
        admin_metadata = volume.admin_metadata

        self.assertEqual(2, len(admin_metadata))
        self.assertEqual('False', admin_metadata['readonly'])
        self.assertEqual('rw', admin_metadata['attached_mode'])
        conn_info = self.volume_api.initialize_connection(self.ctx,
                                                          volume,
                                                          connector)
        self.assertEqual('rw', conn_info['data']['access_mode'])
        # build request to force detach
        volume_remote_error = messaging.RemoteError(exc_type='DBError')
        with mock.patch.object(volume_api.API, 'detach',
                               side_effect=volume_remote_error):
            req = webob.Request.blank('/v3/%s/volumes/%s/action' %
                                      (fake.PROJECT_ID, volume.id))
            req.method = 'POST'
            req.headers['content-type'] = 'application/json'
            body = {'os-force_detach': {'attachment_id': fake.ATTACHMENT_ID,
                                        'connector': connector}}
            req.body = jsonutils.dump_as_bytes(body)
            # attach admin context to request
            req.environ['cinder.context'] = self.ctx
            # make request
            self.assertRaises(messaging.RemoteError,
                              req.get_response,
                              app())

    def test_volume_force_detach_missing_connector(self):
        # current status is available
        volume = self._create_volume(self.ctx, {'provider_location': '',
                                                'size': 1})
        connector = {'initiator': 'iqn.2012-07.org.fake:01'}

        self.volume_api.reserve_volume(self.ctx, volume)
        mountpoint = '/dev/vbd'
        attachment = self.volume_api.attach(self.ctx, volume, fake.INSTANCE_ID,
                                            None, mountpoint, 'rw')
        # volume is attached
        volume.refresh()
        self.assertEqual('in-use', volume.status)
        self.assertEqual(fake.INSTANCE_ID, attachment['instance_uuid'])
        self.assertEqual(mountpoint, attachment['mountpoint'])
        self.assertEqual(fields.VolumeAttachStatus.ATTACHED,
                         attachment['attach_status'])
        admin_metadata = volume.admin_metadata
        self.assertEqual(2, len(admin_metadata))
        self.assertEqual('False', admin_metadata['readonly'])
        self.assertEqual('rw', admin_metadata['attached_mode'])
        conn_info = self.volume_api.initialize_connection(self.ctx,
                                                          volume,
                                                          connector)
        self.assertEqual('rw', conn_info['data']['access_mode'])

        # test when missing connector
        with mock.patch.object(volume_api.API, 'detach'):
            req = webob.Request.blank('/v3/%s/volumes/%s/action' % (
                fake.PROJECT_ID, volume.id))
            req.method = 'POST'
            req.headers['content-type'] = 'application/json'
            body = {'os-force_detach': {'attachment_id': fake.ATTACHMENT_ID}}
            req.body = jsonutils.dump_as_bytes(body)
            # attach admin context to request
            req.environ['cinder.context'] = self.ctx
            # make request
            resp = req.get_response(app())
            self.assertEqual(HTTPStatus.ACCEPTED, resp.status_int)

    def test_attach_in_used_volume_by_instance(self):
        """Test that attaching to an in-use volume fails."""
        # current status is available
        volume = self._create_volume(self.ctx, {'provider_location': '',
                                                'size': 1})
        connector = {'initiator': 'iqn.2012-07.org.fake:01'}
        self.volume_api.reserve_volume(self.ctx, volume)
        conn_info = self.volume_api.initialize_connection(self.ctx,
                                                          volume, connector)
        self.volume_api.attach(self.ctx, volume, fake.INSTANCE_ID, None,
                               '/dev/vbd0', 'rw')
        self.assertEqual('rw', conn_info['data']['access_mode'])
        self.assertRaises(exception.InvalidVolume,
                          self.volume_api.attach,
                          self.ctx,
                          volume,
                          fake.INSTANCE_ID,
                          None,
                          '/dev/vdb1',
                          'ro')

    def test_attach_in_used_volume_by_host(self):
        """Test that attaching to an in-use volume fails."""
        # current status is available
        volume = self._create_volume(self.ctx, {'provider_location': '',
                                                'size': 1})
        connector = {'initiator': 'iqn.2012-07.org.fake:01'}

        self.volume_api.reserve_volume(self.ctx, volume)
        self.volume_api.initialize_connection(self.ctx, volume, connector)
        self.volume_api.attach(self.ctx, volume, None, 'fake_host1',
                               '/dev/vbd0', 'rw')
        conn_info = self.volume_api.initialize_connection(self.ctx,
                                                          volume, connector)
        conn_info['data']['access_mode'] = 'rw'
        self.assertRaises(exception.InvalidVolume,
                          self.volume_api.attach,
                          self.ctx,
                          volume,
                          None,
                          'fake_host2',
                          '/dev/vbd1',
                          'ro')

    def test_invalid_iscsi_connector(self):
        """Test connector without the initiator (required by iscsi driver)."""
        # current status is available
        volume = self._create_volume(self.ctx, {'provider_location': '',
                                                'size': 1})
        connector = {}

        self.assertRaises(exception.InvalidInput,
                          self.volume_api.initialize_connection,
                          self.ctx, volume, connector)

    def test_attach_attaching_volume_with_different_instance(self):
        """Test that attaching volume reserved for another instance fails."""
        # current status is available
        volume = self._create_volume(self.ctx, {'provider_location': '',
                                                'size': 1})

        self.volume_api.reserve_volume(self.ctx, volume)
        values = {'volume_id': volume['id'],
                  'attach_status': fields.VolumeAttachStatus.ATTACHING,
                  'attach_time': timeutils.utcnow(),
                  'instance_uuid': 'abc123',
                  }
        db.volume_attach(self.ctx, values)
        db.volume_admin_metadata_update(self.ctx, volume['id'],
                                        {"attached_mode": 'rw'}, False)
        mountpoint = '/dev/vbd'
        attachment = self.volume_api.attach(self.ctx, volume,
                                            fake.INSTANCE_ID, None,
                                            mountpoint, 'rw')

        self.assertEqual(fake.INSTANCE_ID, attachment['instance_uuid'])
        self.assertEqual(volume['id'], attachment['volume_id'], volume['id'])
        self.assertEqual(fields.VolumeAttachStatus.ATTACHED,
                         attachment['attach_status'])

    def test_attach_attaching_volume_with_different_mode(self):
        """Test that attaching volume reserved for another mode fails."""
        # current status is available
        volume = self._create_volume(
            self.ctx,
            {'provider_location': '',
             'size': 1,
             'status': 'attaching',
             'instance_uuid': fake.INSTANCE_ID,
             'admin_metadata': {"attached_mode": 'rw'}})

        values = {'status': 'attaching'}
        db.volume_update(self.ctx, volume['id'], values)
        db.volume_admin_metadata_update(self.ctx, volume['id'],
                                        {"attached_mode": 'rw'}, False)
        mountpoint = '/dev/vbd'
        self.assertRaises(exception.InvalidVolume,
                          self.volume_api.attach,
                          self.ctx,
                          volume,
                          fake.INSTANCE_ID,
                          None,
                          mountpoint,
                          'ro')



# Copyright (c) 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from cinder.objects import base


@base.CinderObjectRegistry.register
class LogLevel(base.CinderObject):
    """Versioned Object to send log change requests."""
    # Version 1.0: Initial version
    VERSION = '1.0'

    # NOTE: When adding a field obj_make_compatible needs to be updated
    fields = {
        'prefix': fields.StringField(nullable=True),
        'level': fields.StringField(nullable=True),
    }

    def __init__(self, context=None, **kwargs):
        super(LogLevel, self).__init__(**kwargs)

        # Set non initialized fields with default or None values
        for field_name in self.fields:
            if not self.obj_attr_is_set(field_name):
                field = self.fields[field_name]
                if field.default != fields.UnspecifiedDefault:
                    setattr(self, field_name, field.default)
                elif field.nullable:
                    setattr(self, field_name, None)


@base.CinderObjectRegistry.register
class LogLevelList(base.ObjectListBase, base.CinderObject):
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('LogLevel'),
    }



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
import uuid

import paramiko

from cinder import exception
from cinder import ssh_utils
from cinder.tests.unit import test


class FakeSock(object):
    def settimeout(self, timeout):
        pass


class FakeTransport(object):

    def __init__(self):
        self.active = True
        self.sock = FakeSock()

    def set_keepalive(self, timeout):
        pass

    def is_active(self):
        return self.active


class FakeSSHClient(object):

    def __init__(self):
        self.id = uuid.uuid4()
        self.transport = FakeTransport()

    def set_missing_host_key_policy(self, policy):
        self.policy = policy

    def load_system_host_keys(self):
        self.system_host_keys = 'system_host_keys'

    def load_host_keys(self, hosts_key_file):
        self.hosts_key_file = hosts_key_file

    def connect(self, ip, port=22, username=None, password=None,
                pkey=None, timeout=10):
        pass

    def get_transport(self):
        return self.transport

    def get_policy(self):
        return self.policy

    def get_host_keys(self):
        return '127.0.0.1 ssh-rsa deadbeef'

    def close(self):
        pass

    def __call__(self, *args, **kwargs):
        pass


class SSHPoolTestCase(test.TestCase):
    """Unit test for SSH Connection Pool."""
    @mock.patch('paramiko.SSHClient')
    @mock.patch('os.path.isfile', return_value=True)
    def test_sshpool_remove(self, mock_isfile, mock_sshclient):
        ssh_to_remove = mock.MagicMock()
        mock_sshclient.side_effect = [mock.MagicMock(),
                                      ssh_to_remove, mock.MagicMock()]
        self.override_config('ssh_hosts_key_file', 'dummy')
        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=3,
                                    max_size=3)
        self.assertIn(ssh_to_remove, list(sshpool.free_items))
        sshpool.remove(ssh_to_remove)
        self.assertNotIn(ssh_to_remove, list(sshpool.free_items))

    @mock.patch('paramiko.SSHClient')
    @mock.patch('os.path.isfile', return_value=True)
    def test_sshpool_remove_object_not_in_pool(self, mock_isfile,
                                               mock_sshclient):
        # create an SSH Client that is not a part of sshpool.
        ssh_to_remove = mock.MagicMock()
        mock_sshclient.side_effect = [mock.MagicMock(), mock.MagicMock()]

        self.override_config('ssh_hosts_key_file', 'dummy')
        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=2,
                                    max_size=2)
        listBefore = list(sshpool.free_items)
        self.assertNotIn(ssh_to_remove, listBefore)
        sshpool.remove(ssh_to_remove)
        self.assertEqual(listBefore, list(sshpool.free_items))

    @mock.patch('paramiko.SSHClient')
    @mock.patch('os.path.isfile', return_value=True)
    def test_ssh_default_hosts_key_file(self, mock_isfile, mock_sshclient):
        mock_ssh = mock.MagicMock()
        mock_sshclient.return_value = mock_ssh
        self.override_config('ssh_hosts_key_file',
                             '/var/lib/cinder/ssh_known_hosts')

        # create with customized setting
        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=1,
                                    max_size=1)

        host_key_files = sshpool.hosts_key_file

        self.assertEqual('/var/lib/cinder/ssh_known_hosts', host_key_files)

        mock_ssh.load_host_keys.assert_called_once_with(
            '/var/lib/cinder/ssh_known_hosts')

    @mock.patch('paramiko.SSHClient')
    @mock.patch('os.path.isfile', return_value=True)
    def test_ssh_host_key_file_kwargs(self, mock_isfile, mock_sshclient):
        mock_ssh = mock.MagicMock()
        mock_sshclient.return_value = mock_ssh
        self.override_config('ssh_hosts_key_file',
                             '/var/lib/cinder/ssh_known_hosts')

        # create with customized setting
        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=1,
                                    max_size=1,
                                    hosts_key_file='dummy_host_keyfile')

        host_key_files = sshpool.hosts_key_file

        self.assertIn('dummy_host_keyfile', host_key_files)
        self.assertIn('/var/lib/cinder/ssh_known_hosts', host_key_files)

        expected = [
            mock.call.load_host_keys('dummy_host_keyfile'),
            mock.call.load_host_keys('/var/lib/cinder/ssh_known_hosts')]

        mock_ssh.assert_has_calls(expected, any_order=True)

    @mock.patch('os.path.isfile', return_value=True)
    @mock.patch('paramiko.RSAKey.from_private_key_file')
    @mock.patch('paramiko.SSHClient')
    def test_single_ssh_connect(self, mock_sshclient, mock_pkey, mock_isfile):
        self.override_config(
            'ssh_hosts_key_file', '/var/lib/cinder/ssh_known_hosts')

        # create with password
        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=1,
                                    max_size=1)
        with sshpool.item() as ssh:
            first_id = ssh.id

        with sshpool.item() as ssh:
            second_id = ssh.id

        self.assertEqual(first_id, second_id)
        self.assertEqual(1, mock_sshclient.return_value.connect.call_count)

        # create with private key
        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    privatekey="test",
                                    min_size=1,
                                    max_size=1)
        self.assertEqual(2, mock_sshclient.return_value.connect.call_count)

        # attempt to create with no password or private key
        self.assertRaises(paramiko.SSHException,
                          ssh_utils.SSHPool,
                          "127.0.0.1", 22, 10,
                          "test",
                          min_size=1,
                          max_size=1)

    @mock.patch('os.path.isfile', return_value=True)
    @mock.patch('paramiko.SSHClient')
    def test_closed_reopened_ssh_connections(self, mock_sshclient,
                                             mock_open):
        mock_sshclient.return_value = FakeSSHClient()
        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=1,
                                    max_size=4)
        with sshpool.item() as ssh:
            mock_sshclient.reset_mock()
            first_id = ssh.id

        with sshpool.item() as ssh:
            second_id = ssh.id
            ssh.get_transport().active = False
            sshpool.remove(ssh)

        self.assertEqual(first_id, second_id)

        # create a new client
        mock_sshclient.return_value = FakeSSHClient()
        with sshpool.item() as ssh:
            third_id = ssh.id

        self.assertNotEqual(first_id, third_id)

    @mock.patch('paramiko.SSHClient')
    def test_missing_ssh_hosts_key_config(self, mock_sshclient):
        mock_sshclient.return_value = FakeSSHClient()
        self.override_config('ssh_hosts_key_file', None)
        # create with password
        self.assertRaises(exception.ParameterNotFound,
                          ssh_utils.SSHPool,
                          "127.0.0.1", 22, 10,
                          "test",
                          password="test",
                          min_size=1,
                          max_size=1)

    @mock.patch('cinder.ssh_utils.open', new_callable=mock.mock_open)
    @mock.patch('paramiko.SSHClient')
    def test_create_default_known_hosts_file(self, mock_sshclient,
                                             mock_open):
        mock_sshclient.return_value = FakeSSHClient()

        self.flags(state_path='/var/lib/cinder',
                   ssh_hosts_key_file='/var/lib/cinder/ssh_known_hosts')

        default_file = '/var/lib/cinder/ssh_known_hosts'

        ssh_pool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                     "test",
                                     password="test",
                                     min_size=1,
                                     max_size=1)

        with ssh_pool.item() as ssh:
            mock_open.assert_called_once_with(default_file, 'a')
            ssh_pool.remove(ssh)

    @mock.patch('os.path.isfile', return_value=False)
    @mock.patch('paramiko.SSHClient')
    def test_ssh_missing_hosts_key_file(self, mock_sshclient,
                                        mock_isfile):
        mock_sshclient.return_value = FakeSSHClient()

        self.flags(state_path='/var/lib/cinder',
                   ssh_hosts_key_file='/tmp/blah')

        self.assertRaises(exception.InvalidInput,
                          ssh_utils.SSHPool,
                          "127.0.0.1", 22, 10,
                          "test",
                          password="test",
                          min_size=1,
                          max_size=1)

    @mock.patch('paramiko.SSHClient')
    @mock.patch('os.path.isfile', return_value=True)
    def test_ssh_strict_host_key_policy(self, mock_isfile, mock_sshclient):
        mock_sshclient.return_value = FakeSSHClient()

        self.flags(strict_ssh_host_key_policy=True,
                   ssh_hosts_key_file='/var/lib/cinder/ssh_known_hosts')

        # create with customized setting
        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=1,
                                    max_size=1)

        with sshpool.item() as ssh:
            self.assertIsInstance(ssh.get_policy(),
                                  paramiko.RejectPolicy)

    @mock.patch('paramiko.SSHClient')
    @mock.patch('os.path.isfile', return_value=True)
    def test_ssh_not_strict_host_key_policy(self, mock_isfile, mock_sshclient):
        mock_sshclient.return_value = FakeSSHClient()

        self.override_config('strict_ssh_host_key_policy', False)

        # create with customized setting
        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=1,
                                    max_size=1)

        with sshpool.item() as ssh:
            self.assertIsInstance(ssh.get_policy(),
                                  paramiko.AutoAddPolicy)

    @mock.patch('paramiko.SSHClient')
    @mock.patch('cinder.ssh_utils.open', new_callable=mock.mock_open)
    @mock.patch('os.path.isfile', return_value=False)
    def test_ssh_timeout(self, mock_isfile, mock_open, mock_sshclient):
        self.flags(state_path='/var/lib/cinder',
                   ssh_hosts_key_file='/var/lib/cinder/ssh_known_hosts')
        default_file = '/var/lib/cinder/ssh_known_hosts'

        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=1,
                                    max_size=1)
        mock_open.assert_called_once_with(default_file, 'a')
        self.assertEqual(1, sshpool.current_size)
        conn = sshpool.get()
        conn.connect = mock.MagicMock()
        # create failed due to time out
        conn.connect.side_effect = paramiko.SSHException("time out")
        mock_transport = mock.MagicMock()
        conn.get_transport.return_value = mock_transport
        # connection is down
        mock_transport.is_active.return_value = False
        sshpool.put(conn)
        self.assertRaises(paramiko.SSHException,
                          sshpool.get)
        self.assertEqual(0, sshpool.current_size)

    @mock.patch('os.path.isfile', return_value=True)
    @mock.patch('paramiko.RSAKey.from_private_key_file')
    @mock.patch('paramiko.SSHClient')
    def test_ssh_put(self, mock_sshclient, mock_pkey, mock_isfile):
        self.override_config(
            'ssh_hosts_key_file', '/var/lib/cinder/ssh_known_hosts')

        fake_close = mock.MagicMock()
        fake = FakeSSHClient()
        fake.close = fake_close
        mock_sshclient.return_value = fake

        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=5,
                                    max_size=5)
        self.assertEqual(5, sshpool.current_size)
        with sshpool.item():
            pass
        self.assertEqual(5, sshpool.current_size)
        sshpool.resize(4)
        with sshpool.item():
            pass
        self.assertEqual(4, sshpool.current_size)
        fake_close.assert_called_once_with()
        fake_close.reset_mock()
        sshpool.resize(3)
        with sshpool.item():
            pass
        self.assertEqual(3, sshpool.current_size)
        fake_close.assert_called_once_with()

    @mock.patch('os.path.isfile', return_value=True)
    @mock.patch('paramiko.RSAKey.from_private_key_file')
    @mock.patch('paramiko.SSHClient')
    def test_ssh_destructor(self, mock_sshclient, mock_pkey, mock_isfile):
        self.override_config(
            'ssh_hosts_key_file', '/var/lib/cinder/ssh_known_hosts')

        fake_close = mock.MagicMock()
        fake = FakeSSHClient()
        fake.close = fake_close
        mock_sshclient.return_value = fake

        # create with password
        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=5,
                                    max_size=5)
        self.assertEqual(5, sshpool.current_size)
        close_expect_calls = [mock.call(), mock.call(), mock.call(),
                              mock.call(), mock.call()]

        sshpool = ssh_utils.SSHPool("127.0.0.1", 22, 10,
                                    "test",
                                    password="test",
                                    min_size=5,
                                    max_size=5)
        self.assertEqual(fake_close.mock_calls, close_expect_calls)
        sshpool = None
        self.assertEqual(fake_close.mock_calls, close_expect_calls +
                         close_expect_calls)



# Copyright (c) 2016 EMC Corporation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Built-in group type properties."""


from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
import webob

from cinder import context
from cinder import db
from cinder import exception
from cinder.i18n import _

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
DEFAULT_CGSNAPSHOT_TYPE = "default_cgsnapshot_type"


def create(context,
           name,
           group_specs=None,
           is_public=True,
           projects=None,
           description=None):
    """Creates group types."""
    group_specs = group_specs or {}
    projects = projects or []
    elevated = context if context.is_admin else context.elevated()
    try:
        type_ref = db.group_type_create(elevated,
                                        dict(name=name,
                                             group_specs=group_specs,
                                             is_public=is_public,
                                             description=description),
                                        projects=projects)
    except db_exc.DBError:
        LOG.exception('DB error:')
        raise exception.GroupTypeCreateFailed(name=name,
                                              group_specs=group_specs)
    return type_ref


def update(context, id, name, description, is_public=None):
    """Update group type by id."""
    if id is None:
        msg = _("id cannot be None")
        raise exception.InvalidGroupType(reason=msg)
    elevated = context if context.is_admin else context.elevated()
    try:
        db.group_type_update(elevated, id,
                             dict(name=name, description=description,
                                  is_public=is_public))
    except db_exc.DBError:
        LOG.exception('DB error:')
        raise exception.GroupTypeUpdateFailed(id=id)


def destroy(context, id):
    """Marks group types as deleted."""
    if id is None:
        msg = _("id cannot be None")
        raise exception.InvalidGroupType(reason=msg)
    else:
        elevated = context if context.is_admin else context.elevated()
        try:
            db.group_type_destroy(elevated, id)
        except exception.GroupTypeInUse as e:
            msg = _('Target group type is still in use. %s') % e
            raise webob.exc.HTTPBadRequest(explanation=msg)


def get_all_group_types(context, inactive=0, filters=None, marker=None,
                        limit=None, sort_keys=None, sort_dirs=None,
                        offset=None, list_result=False):
    """Get all non-deleted group_types.

    Pass true as argument if you want deleted group types returned also.

    """
    grp_types = db.group_type_get_all(context, inactive, filters=filters,
                                      marker=marker, limit=limit,
                                      sort_keys=sort_keys,
                                      sort_dirs=sort_dirs, offset=offset,
                                      list_result=list_result)
    return grp_types


def get_group_type(ctxt, id, expected_fields=None):
    """Retrieves single group type by id."""
    if id is None:
        msg = _("id cannot be None")
        raise exception.InvalidGroupType(reason=msg)

    if ctxt is None:
        ctxt = context.get_admin_context()

    return db.group_type_get(ctxt, id, expected_fields=expected_fields)


def get_group_type_by_name(context, name):
    """Retrieves single group type by name."""
    if name is None:
        msg = _("name cannot be None")
        raise exception.InvalidGroupType(reason=msg)

    return db.group_type_get_by_name(context, name)


def get_default_group_type():
    """Get the default group type."""
    name = CONF.default_group_type
    grp_type = {}

    if name is not None:
        ctxt = context.get_admin_context()
        try:
            grp_type = get_group_type_by_name(ctxt, name)
        except exception.GroupTypeNotFoundByName:
            # Couldn't find group type with the name in default_group_type
            # flag, record this issue and move on
            LOG.exception('Default group type is not found. '
                          'Please check default_group_type config.')

    return grp_type


def get_default_cgsnapshot_type():
    """Get the default group type for migrating cgsnapshots.

    Get the default group type for migrating consistencygroups to
    groups and cgsnapshots to group_snapshots.
    """

    grp_type = {}

    ctxt = context.get_admin_context()
    try:
        grp_type = get_group_type_by_name(ctxt, DEFAULT_CGSNAPSHOT_TYPE)
    except exception.GroupTypeNotFoundByName:
        # Couldn't find DEFAULT_CGSNAPSHOT_TYPE group type.
        # Record this issue and move on.
        LOG.exception('Default cgsnapshot type %s is not found.',
                      DEFAULT_CGSNAPSHOT_TYPE)

    return grp_type


def is_default_cgsnapshot_type(group_type_id):
    cgsnap_type = get_default_cgsnapshot_type()
    return cgsnap_type and group_type_id == cgsnap_type['id']


def get_group_type_specs(group_type_id, key=False):
    group_type = get_group_type(context.get_admin_context(),
                                group_type_id)
    group_specs = group_type['group_specs']
    if key:
        if group_specs.get(key):
            return group_specs.get(key)
        else:
            return False
    else:
        return group_specs


def is_public_group_type(context, group_type_id):
    """Return is_public boolean value of group type"""
    group_type = db.group_type_get(context, group_type_id)
    return group_type['is_public']


def add_group_type_access(context, group_type_id, project_id):
    """Add access to group type for project_id."""
    if group_type_id is None:
        msg = _("group_type_id cannot be None")
        raise exception.InvalidGroupType(reason=msg)
    elevated = context if context.is_admin else context.elevated()
    if is_public_group_type(elevated, group_type_id):
        msg = _("Type access modification is not applicable to public group "
                "type.")
        raise exception.InvalidGroupType(reason=msg)
    return db.group_type_access_add(elevated, group_type_id, project_id)


def remove_group_type_access(context, group_type_id, project_id):
    """Remove access to group type for project_id."""
    if group_type_id is None:
        msg = _("group_type_id cannot be None")
        raise exception.InvalidGroupType(reason=msg)
    elevated = context if context.is_admin else context.elevated()
    if is_public_group_type(elevated, group_type_id):
        msg = _("Type access modification is not applicable to public group "
                "type.")
        raise exception.InvalidGroupType(reason=msg)
    return db.group_type_access_remove(elevated, group_type_id, project_id)



# Copyright (c) 2019 SandStone data Technologies Co., Ltd
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Unittest for sds_client."""
from unittest import mock
import uuid

import ddt
from oslo_utils import units

from cinder import exception
from cinder import objects
from cinder.tests.unit import test
from cinder.volume import configuration as config
from cinder.volume.drivers.san import san
from cinder.volume.drivers.sandstone import sds_client
from cinder.volume.drivers.sandstone import sds_driver


class FakeSdsBaseDriver(sds_driver.SdsBaseDriver):
    """Fake sds base driver."""

    def __init__(self):
        """Init conf client pool sds_client."""
        self.configuration = config.Configuration(None)
        self.configuration.append_config_values(sds_driver.sds_opts)
        self.configuration.append_config_values(san.san_opts)
        self.configuration.suppress_requests_ssl_warnings = True
        self.client = None
        self.poolid = 1
        self.VERSION = '1.0'
        self.address = "192.168.200.100"
        self.user = "fake_user"
        self.password = "fake_password"
        self.pool = "fake_pool_name"
        self.iscsi_info = {"iqn.1994-05.com.redhat:899c5f9d15d":
                           "1.1.1.1,1.1.1.2,1.1.1.3"}
        self.default_target_ips = ["1.1.1.1", "1.1.1.2", "1.1.1.3"]
        self.default_chap_info = "1234567891234,123456789123"


@ddt.ddt
class TestSdsBaseDriver(test.TestCase):
    """Testcase sds base driver."""

    def setUp(self):
        """Setup."""
        super(TestSdsBaseDriver, self).setUp()
        self.fake_driver = FakeSdsBaseDriver()
        self.fake_driver.client = sds_client.RestCmd('192.168.200.100',
                                                     'fake_user',
                                                     'fake_password',
                                                     True)

    # @mock.patch.object(sds_client.RestCmd, 'login')
    def test_do_setup(self):
        """Do setup."""
        self.fake_driver.client = sds_client.RestCmd(
            'fake_rest_ip', 'user', 'password', True)
        self.fake_driver.configuration.san_ip = 'fake_rest_ip'
        self.fake_driver.configuration.san_login = 'fake_san_user'
        self.fake_driver.configuration.san_password = 'fake_san_password'
        self.fake_driver.do_setup('context')

    @mock.patch.object(sds_client.RestCmd, 'query_pool_info')
    @mock.patch.object(sds_client.RestCmd, 'get_poolid_from_poolname')
    @mock.patch.object(sds_client.RestCmd, 'login')
    def test_check_for_setup_error(self, mock_login,
                                   mock_get_poolid_from_poolname,
                                   mock_query_pool_info):
        """Test pool status health or not."""
        result1 = [
            {'status': {'progress': 33, 'state': ['degraded'], 'flags': 4},
             'pool_name': 'fake_pool_name', 'used': 1792950890,
             'display_name': 'data', 'replicated_size': 2,
             'storage_policy': '2', 'domain_name': 'sandstone',
             'pool_id': 3, 'min_size': 1, 'erasure_code_profile': '',
             'policy_type': 'replicated', 'rule_id': 1,
             'size': 2},
            {'status': {'progress': 33, 'state': ['degraded'], 'flags': 4},
             'pool_name': 'vms1', 'used': 1792950890,
             'display_name': 'data', 'replicated_size': 2,
             'storage_policy': '2', 'domain_name': 'sandstone',
             'pool_id': 3, 'min_size': 1, 'erasure_code_profile': '',
             'policy_type': 'replicated', 'rule_id': 1,
             'size': 2}]
        result2 = [
            {'status': {'progress': 33, 'state': ['degraded'], 'flags': 4},
             'pool_name': 'vms', 'used': 1792950890,
             'display_name': 'data', 'replicated_size': 2,
             'storage_policy': '2', 'domain_name': 'sandstone',
             'pool_id': 3, 'min_size': 1, 'erasure_code_profile': '',
             'policy_type': 'replicated', 'rule_id': 1,
             'size': 2},
            {'status': {'progress': 33, 'state': ['degraded'], 'flags': 4},
             'pool_name': 'vms1', 'used': 1792950890,
             'display_name': 'data', 'replicated_size': 2,
             'storage_policy': '2', 'domain_name': 'sandstone',
             'pool_id': 3, 'min_size': 1, 'erasure_code_profile': '',
             'policy_type': 'replicated', 'rule_id': 1,
             'size': 2}]

        mock_login.return_value = {"success": 1}
        mock_get_poolid_from_poolname.return_value = (
            {"fake_pool_name": 3})
        mock_query_pool_info.return_value = result1
        retval = self.fake_driver.check_for_setup_error()
        self.assertIsNone(retval)
        mock_query_pool_info.return_value = result2
        try:
            self.fake_driver.check_for_setup_error()
        except Exception as e:
            self.assertEqual(exception.InvalidInput, type(e))

    @mock.patch.object(sds_client.RestCmd, 'query_capacity_info')
    def test__update_volume_stats(self, mock_query_capacity_info):
        """Get cluster capacity."""
        result1 = {
            "capacity_bytes": 2 * units.Gi,
            "free_bytes": units.Gi
        }
        mock_query_capacity_info.return_value = result1
        retval = self.fake_driver._update_volume_stats(
            pool_name="fake_pool_name")
        self.assertDictEqual(
            {"pools": [dict(
                pool_name="fake_pool_name",
                vendor_name = 'SandStone USP',
                driver_version = self.fake_driver.VERSION,
                total_capacity_gb=2.0,
                free_capacity_gb=1.0,
                QoS_support=True,
                thin_provisioning_support=True,
                multiattach=False,)
            ]}, retval)
        mock_query_capacity_info.assert_called_once_with()

    @mock.patch.object(sds_driver.SdsBaseDriver, 'get_volume_stats')
    def test_get_volume_stats(self, mock_get_volume_stats):
        """Get cluster capacitys."""
        result1 = {"pool": dict(
            pool_name="fake_pool_name",
            total_capacity_gb=2.0,
            free_capacity_gb=1.0,
            QoS_support=True,
            thin_provisioning_support=True,
            multiattach=False,)}
        mock_get_volume_stats.return_value = result1
        retval = self.fake_driver.get_volume_stats()
        self.assertDictEqual(
            {"pool": dict(
                pool_name="fake_pool_name",
                total_capacity_gb=2.0,
                free_capacity_gb=1.0,
                QoS_support=True,
                thin_provisioning_support=True,
                multiattach=False,
            )}, retval)

    @mock.patch.object(sds_client.RestCmd, 'create_lun')
    def test_create_volume(self, mock_create_lun):
        """Test create volume."""
        volume = objects.Volume(_name_id=uuid.uuid4(), size=1)
        mock_create_lun.return_value = {'success': 1}
        retval = self.fake_driver.create_volume(volume=volume)
        self.assertIsNone(retval)

    @mock.patch.object(sds_client.RestCmd, 'delete_lun')
    def test_delete_volume(self, mock_delete_):
        """Test delete volume."""
        mock_delete_.return_value = {'success': 1}
        volume = objects.Volume(_name_id=uuid.uuid4(), size=1)
        retval = self.fake_driver.delete_volume(volume)
        self.assertIsNone(retval)

    @mock.patch.object(sds_client.RestCmd, 'extend_lun')
    @mock.patch.object(sds_client.RestCmd, 'create_lun_from_snapshot')
    def test_create_volume_from_snapshot(self, mock_lun_from_snapshot,
                                         mock_extend_lun):
        """Test create new volume from snapshot of src volume."""
        volume = objects.Volume(_name_id=uuid.uuid4(), size=1)
        snapshot = objects.Snapshot(
            id=uuid.uuid4(), volume_size=2, volume=volume)
        mock_lun_from_snapshot.return_value = {'success': 1}
        mock_extend_lun.return_value = {'success': 1}
        retval = self.fake_driver.create_volume_from_snapshot(volume, snapshot)
        self.assertIsNone(retval)

    @mock.patch.object(sds_client.RestCmd, 'extend_lun')
    @mock.patch.object(sds_client.RestCmd, 'create_lun_from_lun')
    @mock.patch.object(sds_driver.SdsBaseDriver, '_check_volume_exist')
    def test_create_cloned_volume(self, mock__check_volume_exist,
                                  mock_create_lun_from_lun,
                                  mock_extend_lun):
        """Test create clone volume."""
        mock__check_volume_exist.return_value = True
        mock_create_lun_from_lun.return_value = {'success': 1}
        mock_extend_lun.return_value = {'success': 1}
        dst_volume = objects.Volume(_name_id=uuid.uuid4(), size=2)
        src_volume = objects.Volume(_name_id=uuid.uuid4(), size=1)
        retval = self.fake_driver.create_cloned_volume(dst_volume, src_volume)
        self.assertIsNone(retval)

    @mock.patch.object(sds_client.RestCmd, 'query_lun_by_name')
    def test__check_volume_exist(self, mock_query_lun_by_name):
        """Test volume exist or not."""
        mock_query_lun_by_name.return_value = {'success': 1}
        volume = objects.Volume(_name_id=uuid.uuid4(), size=1)
        retval = self.fake_driver._check_volume_exist(volume)
        self.assertEqual({'success': 1}, retval)

    @mock.patch.object(sds_client.RestCmd, 'extend_lun')
    @mock.patch.object(sds_driver.SdsBaseDriver, '_check_volume_exist')
    def test_extend_volume(self, mock__check_volume_exist, mock_extend_lun):
        """Test resize volume."""
        volume = objects.Volume(_name_id=uuid.uuid4(), size=1)
        new_size = 3
        mock__check_volume_exist.return_value = {
            'capacity_bytes': units.Gi * 1}
        mock_extend_lun.return_value = {'success': 1}
        retval = self.fake_driver.extend_volume(volume, new_size)
        self.assertIsNone(retval)

    @mock.patch.object(sds_client.RestCmd, 'create_snapshot')
    def test_create_snapshot(self, mock_create_snapshot):
        """Test create snapshot of volume."""
        volume = objects.Volume(_name_id=uuid.uuid4(), size=1)
        snapshot = objects.Snapshot(
            id=uuid.uuid4(), volume_size=2, volume=volume)
        mock_create_snapshot.return_value = {'success': 1}
        retval = self.fake_driver.create_snapshot(snapshot)
        self.assertIsNone(retval)

    @mock.patch.object(sds_client.RestCmd, 'query_snapshot_by_name')
    def test__check_snapshot_exist(self, mock_query_snapshot_by_name):
        """Test snapshot exist or not."""
        volume = objects.Volume(_name_id=uuid.uuid4(), size=1)
        snapshot = objects.Snapshot(
            id=uuid.uuid4(), volume_size=2, volume=volume)
        mock_query_snapshot_by_name.return_value = {'success': 1}
        retval = self.fake_driver._check_snapshot_exist(snapshot)
        self.assertEqual({'success': 1}, retval)

    @mock.patch.object(sds_client.RestCmd, 'delete_snapshot')
    @mock.patch.object(sds_driver.SdsBaseDriver, '_check_snapshot_exist')
    def test_delete_snapshot(self, mock__check_snapshot_exist,
                             mock_delete_snapshot):
        """Test delete snapshot."""
        volume = objects.Volume(_name_id=uuid.uuid4(), size=1)
        snapshot = objects.Snapshot(
            id=uuid.uuid4(), volume_size=2, volume=volume)
        mock__check_snapshot_exist.return_value = True
        mock_delete_snapshot.return_value = {'success': 1}
        retval = self.fake_driver.delete_snapshot(snapshot)
        self.assertIsNone(retval)


class FakeSdsISCSIDriver(sds_driver.SdsISCSIDriver):
    """Fake sds iscsi driver, include attach, detach."""

    def __init__(self):
        """Init conf client pool."""
        self.configuration = config.Configuration(None)
        self.client = None
        self.address = "192.168.200.100"
        self.user = "fake_user"
        self.password = "fake_password"
        self.pool = "fake_pool_name"
        self.poolid = 1
        self.iscsi_info = {"iqn.1994-05.com.redhat:899c5f9d15d":
                           "1.1.1.1,1.1.1.2,1.1.1.3"}
        self.default_target_ips = ["1.1.1.1", "1.1.1.2", "1.1.1.3"]
        self.chap_username = "123456789123"
        self.chap_password = "1234567891234"


@ddt.ddt
class TestSdsISCSIDriver(test.TestCase):
    """Testcase sds iscsi driver, include attach, detach."""

    def setUp(self):
        """Setup."""
        super(TestSdsISCSIDriver, self).setUp()
        self.fake_driver = FakeSdsISCSIDriver()
        self.fake_driver.client = sds_client.RestCmd("192.168.200.100",
                                                     "fake_user",
                                                     "fake_password",
                                                     True)

    @mock.patch.object(sds_client.RestCmd, 'query_target_by_name')
    def test__check_target_exist(self, mock_query_target_by_name):
        """Test target exist or not."""
        target_name = 'test_driver'
        mock_query_target_by_name.return_value = {'success': 1}
        retval = self.fake_driver._check_target_exist(target_name)
        self.assertEqual({'success': 1}, retval)

    @mock.patch.object(sds_client.RestCmd, 'query_initiator_by_name')
    def test__check_initiator_exist(self, mock_query_initiator_by_name):
        """Test initiator exist or not."""
        initiator_name = 'test_driver'
        mock_query_initiator_by_name.return_value = {'success': 1}
        retval = self.fake_driver._check_initiator_exist(initiator_name)
        self.assertEqual({'success': 1}, retval)

    @mock.patch.object(sds_client.RestCmd, 'query_target_initiatoracl')
    def test__check_target_added_initiator(self,
                                           mock_query_target_initiatoracl):
        """Test target added the initiator."""
        mock_query_target_initiatoracl.return_value = {'success': 1}
        target_name, initiator_name = 'test_driver', 'initiator_name'
        retval = self.fake_driver._check_target_added_initiator(target_name,
                                                                initiator_name)
        self.assertEqual({'success': 1}, retval)

    @mock.patch.object(sds_client.RestCmd, 'query_target_lunacl')
    def test__check_target_added_lun(self, mock_query_target_lunacl):
        """Test target added the lun."""
        mock_query_target_lunacl.return_value = {'success': 1}
        target_name, pool_name, volume_name = ('ccc', self.fake_driver.pool,
                                               'fcc')
        retval = self.fake_driver._check_target_added_lun(target_name,
                                                          pool_name,
                                                          volume_name)
        self.assertEqual({'success': 1}, retval)

    @mock.patch.object(sds_client.RestCmd, 'query_chapinfo_by_target')
    def test__check_target_added_chap(self, mock_query_chapinfo_by_target):
        """Test target added chapuser."""
        mock_query_chapinfo_by_target.return_value = {'success': 1}
        target_name, user_name = 'ccc', 'fcc'
        retval = self.fake_driver._check_target_added_chap(target_name,
                                                           user_name)
        self.assertEqual({'success': 1}, retval)

    def test__get_target_ip(self):
        """Test get target from targetip."""
        initiator = 'iqn.1994-05.com.redhat:899c5f9d15d'
        retval_target_ips = \
            self.fake_driver._get_target_ip(initiator)
        self.assertListEqual(['1.1.1.1', '1.1.1.2', '1.1.1.3'],
                             retval_target_ips)

        self.fake_driver.default_target_ips = \
            ["1.1.1.1"]
        initiator = 'vms'
        retval_target_ips = \
            self.fake_driver._get_target_ip(initiator)
        self.assertListEqual(["1.1.1.1"], retval_target_ips)

    @mock.patch.object(sds_client.RestCmd, 'add_chap_by_target')
    @mock.patch.object(sds_driver.SdsISCSIDriver, '_check_target_added_chap')
    @mock.patch.object(sds_driver.SdsISCSIDriver, '_check_target_added_lun')
    @mock.patch.object(sds_client.RestCmd, 'mapping_lun')
    @mock.patch.object(sds_client.RestCmd, 'add_initiator_to_target')
    @mock.patch.object(sds_driver.SdsISCSIDriver,
                       '_check_target_added_initiator')
    @mock.patch.object(sds_client.RestCmd, 'create_initiator')
    @mock.patch.object(sds_driver.SdsISCSIDriver, '_check_initiator_exist')
    @mock.patch.object(sds_client.RestCmd, 'create_target')
    @mock.patch.object(sds_client.RestCmd, 'query_node_by_targetips')
    @mock.patch.object(sds_driver.SdsISCSIDriver, '_check_target_exist')
    @mock.patch.object(sds_driver.SdsISCSIDriver, '_get_target_ip')
    def test_initialize_connection(self, mock__get_target_ip,
                                   mock__check_target_exist,
                                   mock_query_node_by_targetips,
                                   mock_create_target,
                                   mock__check_initiator_exist,
                                   mock_create_initiator,
                                   mock__check_target_added_initiator,
                                   mock_add_initiator_to_target,
                                   mock_mapping_lun,
                                   mock__check_target_added_lun,
                                   mock__check_target_added_chap,
                                   mock_add_chap_by_target):
        """Test attach volume to kvm."""
        mock__get_target_ip.return_value = (['1.1.1.1', '1.1.1.2', '1.1.1.3'])
        mock__check_target_exist.return_value = False
        mock__check_initiator_exist.return_value = False
        mock__check_target_added_initiator.result_value = False
        mock__check_target_added_chap.return_value = False
        mock_query_node_by_targetips.return_value = {'host_id', 'address'}
        mock_create_target.return_value = {'success': 1}
        mock_create_initiator.return_value = {'success': 1}
        mock_add_initiator_to_target.result_value = {'success': 1}
        mock_mapping_lun.return_value = {'success': 1}
        mock__check_target_added_lun.return_value = 1
        mock_add_chap_by_target.return_value = {'success': 1}

        volume1, connector1 = (objects.Volume(id=uuid.uuid4(),
                                              _name_id=uuid.uuid4(), size=1),
                               {'initiator':
                                'iqn.1994-05.com.redhat:899c5f9d15d',
                                'multipath': True})
        initiator_name = connector1['initiator']
        iqn_end = initiator_name.split(':', 1)[1]
        target_head = 'iqn.2014-10.com.szsandstone:storage:'
        target_name = target_head + iqn_end
        result1 = {
            'driver_volume_type': 'iscsi',
            'data': {'target_discovered': True,
                     'target_portals': ['1.1.1.1:3260',
                                        '1.1.1.2:3260',
                                        '1.1.1.3:3260'],
                     'volume_id': volume1.id,
                     'auth_method': 'CHAP',
                     'auth_username': '123456789123',
                     'auth_password': '1234567891234',
                     'target_iqns': [target_name, target_name, target_name],
                     'target_luns': [1, 1, 1]}}
        retval = self.fake_driver.initialize_connection(volume1, connector1)
        self.assertDictEqual(result1, retval)

        volume2, connector2 = (objects.Volume(id=uuid.uuid4(),
                                              _name_id=uuid.uuid4(),
                                              size=2),
                               {'initiator':
                                'iqn.1994-05.com.redhat:899c5f9d15d'})
        mock__get_target_ip.return_value = (['1.1.1.1', '1.1.1.2', '1.1.1.3'])
        initiator_name = connector2['initiator']
        iqn_end = initiator_name.split(':', 1)[1]
        target_head = 'iqn.2014-10.com.szsandstone:storage:'
        target_name = target_head + iqn_end
        result2 = {'driver_volume_type': 'iscsi',
                   'data': {'target_discovered': True,
                            'target_portal': '1.1.1.1:3260',
                            'volume_id': volume2.id,
                            'target_iqn': target_name,
                            'target_lun': 1,
                            'auth_method': 'CHAP',
                            'auth_username': '123456789123',
                            'auth_password': '1234567891234'}}
        retval = self.fake_driver.initialize_connection(volume2, connector2)
        self.assertDictEqual(result2, retval)

    @mock.patch.object(sds_client.RestCmd, 'unmap_lun')
    def test_terminate_connection(self, mock_unmap_lun):
        """Test detach volume from kvm."""
        volume, connector = (objects.Volume(_name_id=uuid.uuid4(), size=1),
                             {'initiator':
                              'iqn.1994-05.com.redhat:899c5f9d15d'})
        mock_unmap_lun.result_value = {'success': 1}
        retval = self.fake_driver.terminate_connection(volume, connector)
        self.assertIsNone(retval)



# Copyright 2020 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import HTTPStatus
from unittest import mock
import uuid

import ddt
from webob import exc

from cinder.api import api_utils
from cinder.api import microversions as mv
from cinder.api.v3 import default_types
from cinder import db
from cinder.policies import default_types as default_type_policies
from cinder.tests.unit.api import fakes as fake_api
from cinder.tests.unit import fake_constants
from cinder.tests.unit.policies import base
from cinder.tests.unit.policies import test_base
from cinder.tests.unit import utils as test_utils


class FakeProject(object):
    def __init__(self, id=None, name=None):
        if id:
            self.id = id
        else:
            self.id = uuid.uuid4().hex
        self.name = name
        self.description = 'fake project description'
        self.domain_id = 'default'


class DefaultVolumeTypesPolicyTests(test_base.CinderPolicyTests):

    class FakeDefaultType:
        project_id = fake_constants.PROJECT_ID
        volume_type_id = fake_constants.VOLUME_TYPE_ID

    def setUp(self):
        super(DefaultVolumeTypesPolicyTests, self).setUp()
        self.volume_type = self._create_fake_type(self.admin_context)
        self.project = FakeProject()
        # Need to mock out Keystone so the functional tests don't require other
        # services
        _keystone_client = mock.MagicMock()
        _keystone_client.version = 'v3'
        _keystone_client.projects.get.side_effect = self._get_project
        _keystone_client_get = mock.patch(
            'cinder.api.api_utils._keystone_client',
            lambda *args, **kwargs: _keystone_client)
        _keystone_client_get.start()
        self.addCleanup(_keystone_client_get.stop)

    def _get_project(self, project_id, *args, **kwargs):
        return self.project

    def test_system_admin_can_set_default(self):
        system_admin_context = self.system_admin_context

        path = '/v3/default-types/%s' % system_admin_context.project_id
        body = {
            'default_type':
                {"volume_type": self.volume_type.id}
        }
        response = self._get_request_response(system_admin_context,
                                              path, 'PUT', body=body,
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(HTTPStatus.OK, response.status_int)

    def test_project_admin_can_set_default(self):
        admin_context = self.admin_context

        path = '/v3/default-types/%s' % admin_context.project_id
        body = {
            'default_type':
                {"volume_type": self.volume_type.id}
        }
        response = self._get_request_response(admin_context,
                                              path, 'PUT', body=body,
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(HTTPStatus.OK, response.status_int)

    @mock.patch.object(db, 'project_default_volume_type_get',
                       return_value=FakeDefaultType())
    def test_system_admin_can_get_default(self, mock_default_get):
        system_admin_context = self.system_admin_context

        path = '/v3/default-types/%s' % system_admin_context.project_id
        response = self._get_request_response(system_admin_context,
                                              path, 'GET',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(HTTPStatus.OK, response.status_int)

    def test_project_admin_can_get_default(self):
        admin_context = self.admin_context

        path = '/v3/default-types/%s' % admin_context.project_id
        body = {
            'default_type':
                {"volume_type": self.volume_type.id}
        }
        self._get_request_response(admin_context,
                                   path, 'PUT', body=body,
                                   microversion=
                                   mv.DEFAULT_TYPE_OVERRIDES)

        path = '/v3/default-types/%s' % admin_context.project_id
        response = self._get_request_response(admin_context,
                                              path, 'GET',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(HTTPStatus.OK, response.status_int)

    def test_system_admin_can_get_all_default(self):
        system_admin_context = self.system_admin_context

        path = '/v3/default-types'
        response = self._get_request_response(system_admin_context,
                                              path, 'GET',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(HTTPStatus.OK, response.status_int)

    def test_system_admin_can_unset_default(self):
        system_admin_context = self.system_admin_context

        path = '/v3/default-types/%s' % system_admin_context.project_id
        response = self._get_request_response(system_admin_context,
                                              path, 'DELETE',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(HTTPStatus.NO_CONTENT, response.status_int)

    def test_project_admin_can_unset_default(self):
        admin_context = self.admin_context

        path = '/v3/default-types/%s' % admin_context.project_id
        response = self._get_request_response(admin_context,
                                              path, 'DELETE',
                                              microversion=
                                              mv.DEFAULT_TYPE_OVERRIDES)

        self.assertEqual(HTTPStatus.NO_CONTENT, response.status_int)


@ddt.ddt
class DefaultVolumeTypesPolicyTest(base.BasePolicyTest):

    authorized_admins = [
        'system_admin',
        'legacy_admin',
        'project_admin',
    ]
    unauthorized_admins = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_member',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    # Basic policy test is without enforcing scope (which cinder doesn't
    # yet support) and deprecated rules enabled.
    def setUp(self, enforce_scope=False, enforce_new_defaults=False,
              *args, **kwargs):
        super().setUp(enforce_scope, enforce_new_defaults, *args, **kwargs)
        self.controller = default_types.DefaultTypesController()
        self.api_path = '/v3/default-types/%s' % (self.project_id)
        self.api_version = mv.DEFAULT_TYPE_OVERRIDES

    def _create_volume_type(self):
        vol_type = test_utils.create_volume_type(self.project_admin_context,
                                                 name='fake_vol_type',
                                                 testcase_instance=self)
        return vol_type

    @ddt.data(*base.all_users)
    @mock.patch.object(api_utils, 'get_project')
    def test_default_type_set_policy(self, user_id, fake_project):
        vol_type = self._create_volume_type()
        fake_project.return_value = FakeProject(id=self.project_id)
        rule_name = default_type_policies.CREATE_UPDATE_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'POST'
        body = {"default_type": {"volume_type": vol_type.id}}
        unauthorized_exceptions = [exc.HTTPForbidden]
        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.create_update, req,
                                 id=vol_type.id, body=body)

    @ddt.data(*base.all_users)
    @mock.patch.object(default_types.db, 'project_default_volume_type_get')
    @mock.patch.object(api_utils, 'get_project')
    def test_default_type_get_policy(self, user_id, fake_project,
                                     mock_default_get):
        fake_project.return_value = FakeProject(id=self.project_id)
        rule_name = default_type_policies.GET_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        unauthorized_exceptions = [exc.HTTPForbidden]
        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.detail, req,
                                 id=self.project_id)

    @ddt.data(*base.all_users)
    @mock.patch.object(default_types.db, 'project_default_volume_type_get')
    def test_default_type_get_all_policy(self, user_id, mock_default_get):
        rule_name = default_type_policies.GET_ALL_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        unauthorized_exceptions = [exc.HTTPForbidden]
        # NOTE: The users 'legacy_admin' and 'project_admin' pass for
        # GET_ALL_POLICY since with enforce_new_defaults=False, we have
        # a logical OR between old policy and new one hence RULE_ADMIN_API
        # allows them to pass
        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.index, req)

    @ddt.data(*base.all_users)
    @mock.patch.object(api_utils, 'get_project')
    @mock.patch.object(default_types.db, 'project_default_volume_type_get')
    def test_default_type_unset_policy(self, user_id, mock_default_unset,
                                       fake_project):
        fake_project.return_value = FakeProject(id=self.project_id)
        rule_name = default_type_policies.DELETE_POLICY
        url = self.api_path
        req = fake_api.HTTPRequest.blank(url, version=self.api_version)
        req.method = 'DELETE'
        unauthorized_exceptions = [exc.HTTPForbidden]
        self.common_policy_check(user_id, self.authorized_admins,
                                 self.unauthorized_admins,
                                 unauthorized_exceptions,
                                 rule_name, self.controller.delete, req,
                                 id=self.project_id)


class DefaultVolumeTypesPolicySecureRbacTest(DefaultVolumeTypesPolicyTest):

    authorized_admins = [
        'legacy_admin',
        'system_admin',
        'project_admin',
    ]
    unauthorized_admins = [
        'legacy_owner',
        'system_member',
        'system_reader',
        'system_foo',
        'project_member',
        'project_reader',
        'project_foo',
        'other_project_member',
        'other_project_reader',
    ]

    def setUp(self, *args, **kwargs):
        # Test secure RBAC by disabling deprecated policy rules (scope
        # is still not enabled).
        super().setUp(enforce_scope=False, enforce_new_defaults=True,
                      *args, **kwargs)
